# Standard Library Imports
from typing import Any, Dict, List

# RAMSTK Package Imports
from ramstk.configuration import RAMSTKUserConfiguration as RAMSTKUserConfiguration
from ramstk.logger import RAMSTKLogManager as RAMSTKLogManager
from ramstk.views.gtk3 import Gtk as Gtk
from ramstk.views.gtk3 import _ as _
from ramstk.views.gtk3.assistants import AddStressTestMethod as AddStressTestMethod
from ramstk.views.gtk3.widgets import RAMSTKMessageDialog as RAMSTKMessageDialog
from ramstk.views.gtk3.widgets import RAMSTKPanel as RAMSTKPanel
from ramstk.views.gtk3.widgets import RAMSTKWorkView as RAMSTKWorkView

# RAMSTK Local Imports
from . import PoFTreePanel as PoFTreePanel

class PoFWorkView(RAMSTKWorkView):
    _tag: str
    _pixbuf: bool
    _tablabel: str
    _tabtooltip: str
    _lst_tooltips: List[str]
    _pnlPanel: RAMSTKPanel
    def __init__(
        self, configuration: RAMSTKUserConfiguration, logger: RAMSTKLogManager
    ) -> None: ...
    def _do_request_delete(self, __button: Gtk.ToolButton) -> None: ...
    def _do_request_insert_child(self, __button: Gtk.ToolButton) -> None: ...
    def _do_request_insert_sibling(self, __button: Gtk.ToolButton) -> None: ...
    _record_id: int
    def _do_set_record_id(self, attributes: Dict[str, Any]) -> None: ...
    def __do_load_test_method_lists(self) -> None: ...
    def __make_ui(self) -> None: ...
    def __on_request_insert_opstress_method(self) -> str: ...



# -*- coding: utf-8 -*-
#
#       ramstk.views.gtk3.failure_definition.panel.py is part of The RAMSTK Project
#
# All rights reserved.
# Copyright since 2007 Doyle "weibullguy" Rowland doyle.rowland <AT> reliaqual <DOT> com
"""GTK3 Failure Definition Panels."""

# Standard Library Imports
from typing import Any, Dict, List

# Third Party Imports
from pubsub import pub

# RAMSTK Package Imports
from ramstk.views.gtk3 import Gtk, _
from ramstk.views.gtk3.widgets import RAMSTKTreePanel


class FailureDefinitionTreePanel(RAMSTKTreePanel):
    """Panel to display list of failure definitions."""

    # Define private dictionary class attributes.

    # Define private list class attributes.

    # Define private scalar class attributes.
    _select_msg = "succeed_retrieve_all_definition"
    _tag = "definition"
    _title = _("Failure Definition List")

    # Define public dictionary class attributes.

    # Define public list class attributes.

    # Define public scalar class attributes.

    def __init__(self) -> None:
        """Initialize an instance of the failure definition panel."""
        super().__init__()

        # Initialize private dictionary class attributes.
        self.tvwTreeView.dic_row_loader = {
            "definition": super().do_load_treerow,
        }

        # Initialize private list class attributes.

        # Initialize private scalar class attributes.
        self._filtered_tree = True
        self._on_edit_message: str = f"wvw_editing_{self._tag}"

        # Initialize public dictionary class attributes.
        self.dic_attribute_widget_map: Dict[str, List[Any]] = {
            "revision_id": [
                0,
                Gtk.CellRendererText(),
                "edited",
                None,
                self._on_edit_message,
                0,
                {
                    "bg_color": "#FFFFFF",
                    "editable": False,
                    "fg_color": "#000000",
                    "visible": False,
                },
                _("Revision ID"),
                "gint",
            ],
            "function_id": [
                1,
                Gtk.CellRendererText(),
                "edited",
                None,
                self._on_edit_message,
                0,
                {
                    "bg_color": "#FFFFFF",
                    "editable": False,
                    "fg_color": "#000000",
                    "visible": False,
                },
                _("Function ID"),
                "gint",
            ],
            "definition_id": [
                2,
                Gtk.CellRendererText(),
                "edited",
                None,
                self._on_edit_message,
                0,
                {
                    "bg_color": "#FFFFFF",
                    "editable": False,
                    "fg_color": "#000000",
                    "visible": True,
                },
                _("Definition ID"),
                "gint",
            ],
            "definition": [
                3,
                Gtk.CellRendererText(),
                "edited",
                super().on_cell_edit,
                self._on_edit_message,
                "",
                {
                    "bg_color": "#FFFFFF",
                    "editable": True,
                    "fg_color": "#000000",
                    "visible": True,
                },
                _("Definition"),
                "gchararray",
            ],
        }

        # Initialize public list class attributes.

        # Initialize public scalar class attributes.

        super().do_set_properties()
        super().do_make_panel()
        super().do_set_callbacks()

        self.tvwTreeView.set_tooltip_text(
            _("Displays the list of failure definitions for the selected revision.")
        )

        # Subscribe to PyPubSub messages.
        pub.subscribe(self._on_select_function, "selected_function")

    # pylint: disable=unused-argument
    # noinspection PyUnusedLocal
    def do_filter_tree(
        self, model: Gtk.TreeModel, row: Gtk.TreeIter, data: Any
    ) -> bool:
        """Filter Hazards to show only those associated with the selected Function.

        :param model: the filtered model for the Hazard RAMSTKTreeView.
        :param row: the iter to check against condition(s).
        :param data: unused in this method; required by Gtk.TreeModelFilter() widget.
        :return: True if row should be visible, False else.
        :rtype: bool
        """
        return model[row][1] == self._parent_id

    def _on_row_change(self, selection: Gtk.TreeSelection) -> None:
        """Read attributes from newly selected RAMSTKTreeView() row.

        This method is called whenever a view's RAMSTKTreeView() row is
        activated/changed.

        :param selection: the Gtk.TreeSelection() for the newly selected row.
        :return: None
        """
        _attributes = super().on_row_change(selection)

        if _attributes:
            self._record_id = _attributes["definition_id"]

            pub.sendMessage(
                "selected_failure_definition",
                attributes=_attributes,
            )

    def _on_select_function(self, attributes: Dict[str, Any]) -> None:
        """Filter hazards list when Function is selected.

        :param attributes: the dict of Function attributes for the selected Function.
        :return: None
        :rtype: None
        """
        self._parent_id = attributes["function_id"]
        self.tvwTreeView.filt_model.refilter()



# -*- coding: utf-8 -*-
#
#       ramstk.models.dbrecords.commondb_manufacturer_record.py is part of The RAMSTK
#       Project
#
# All rights reserved.
# Copyright since 2007 Doyle "weibullguy" Rowland doyle.rowland <AT> reliaqual <DOT> com
"""Manufacturer Record Model."""

# Standard Library Imports
from typing import Dict, Union

# Third Party Imports
from sqlalchemy import Column, Integer, String

# RAMSTK Local Imports
from .. import RAMSTK_BASE
from .baserecord import RAMSTKBaseRecord


class RAMSTKManufacturerRecord(RAMSTK_BASE, RAMSTKBaseRecord):  # type: ignore
    """Class to represent ramstk_manufacturer in the RAMSTK Common database."""

    __defaults__ = {
        "description": "Manufacturer Description",
        "location": "unknown",
        "cage_code": "CAGE Code",
    }
    __tablename__ = "ramstk_manufacturer"
    __table_args__ = {"extend_existing": True}

    manufacturer_id = Column(
        "fld_manufacturer_id",
        Integer,
        primary_key=True,
        autoincrement=True,
        nullable=False,
    )
    description = Column(
        "fld_description", String(512), default=__defaults__["description"]
    )
    location = Column("fld_location", String(512), default=__defaults__["location"])
    cage_code = Column("fld_cage_code", String(512), default=__defaults__["cage_code"])

    def get_attributes(self) -> Dict[str, Union[int, str]]:
        """Retrieve current values of RAMSTKManufacturer data model attributes.

        :return: {manufacturer_id, description, location, cage_code} pairs
        :rtype: dict
        """
        return {
            "manufacturer_id": self.manufacturer_id,
            "description": self.description,
            "location": self.location,
            "cage_code": self.cage_code,
        }



# -*- coding: utf-8 -*-
#
#       ramstk.models.dbrecords.commondb_stakeholders_record.py is part of The RAMSTK
#       Project
#
# All rights reserved.
# Copyright since 2007 Doyle "weibullguy" Rowland doyle.rowland <AT> reliaqual <DOT> com
"""RAMSTKStakeholders Record Model."""

# Standard Library Imports
from typing import Dict, Union

# Third Party Imports
from sqlalchemy import Column, Integer, String

# RAMSTK Local Imports
from .. import RAMSTK_BASE
from .baserecord import RAMSTKBaseRecord


class RAMSTKStakeholdersRecord(RAMSTK_BASE, RAMSTKBaseRecord):  # type: ignore
    """Class to represent ramstk_stakeholders in the RAMSTK Common database."""

    __defaults__ = {
        "stakeholder": "Stakeholder",
    }
    __tablename__ = "ramstk_stakeholders"
    __table_args__ = {"extend_existing": True}

    stakeholders_id = Column(
        "fld_stakeholders_id",
        Integer,
        primary_key=True,
        autoincrement=True,
        nullable=False,
    )
    stakeholder = Column(
        "fld_stakeholder", String(512), default=__defaults__["stakeholder"]
    )

    def get_attributes(self) -> Dict[str, Union[int, str]]:
        """Retrieve current values of RAMSTKStakeholders data model attributes.

        :return: {stakeholders_id, stakeholder} pairs.
        :rtype: dict
        """
        return {
            "stakeholders_id": self.stakeholders_id,
            "stakeholder": self.stakeholder,
        }



# pylint: disable=unused-import
# -*- coding: utf-8 -*-
#
#       ramstk.models.db.__init__.py is part of The RAMSTK Project
#
# All rights reserved.
# Copyright since 2007 Doyle "weibullguy" Rowland doyle.rowland <AT> reliaqual <DOT> com
"""The RAMSTK database models package."""

# RAMSTK Local Imports
from .basedatabase import BaseDatabase, do_create_program_db
from .common_database import RAMSTKCommonDB
from .program_database import RAMSTKProgramDB



# -*- coding: utf-8 -*-
#
#       ramstk.models.dbtables.programdb_revision_table.py is part of The RAMSTK Project
#
# All rights reserved.
# Copyright since 2007 Doyle "weibullguy" Rowland doyle.rowland <AT> reliaqual <DOT> com
"""RAMSTKRevision Table Model."""

# Standard Library Imports
from datetime import date
from typing import Dict, Type, Union

# RAMSTK Local Imports
from ..dbrecords import RAMSTKRevisionRecord
from .basetable import RAMSTKBaseTable


class RAMSTKRevisionTable(RAMSTKBaseTable):
    """Contain the attributes and methods of the Revision table model."""

    # Define private dictionary class attributes.

    # Define private list class attributes.

    # Define private scalar class attributes.
    _db_id_colname = "fld_revision_id"
    _db_tablename = "ramstk_revision"
    _select_msg = "request_retrieve_revisions"
    _tag = "revision"

    # Define public dictionary class attributes.

    # Define public list class attributes.

    # Define public scalar class attributes.

    def __init__(self, **kwargs: Dict[str, Union[float, int, str]]) -> None:
        """Initialize a RAMSTKRevision table model instance."""
        super().__init__(**kwargs)

        # Initialize private dictionary attributes.

        # Initialize private list attributes.
        self._lst_id_columns = [
            "revision_id",
        ]

        # Initialize private scalar attributes.
        self._record: Type[RAMSTKRevisionRecord] = RAMSTKRevisionRecord

        # Initialize public dictionary attributes.

        # Initialize public list attributes.

        # Initialize public scalar attributes.
        self.pkey = "revision_id"

        # Subscribe to PyPubSub messages.

    # pylint: disable=unused-argument, disable=method-hidden
    def do_get_new_record(
        self, attributes: Dict[str, Union[date, float, int, str]]
    ) -> RAMSTKRevisionRecord:
        """Gets a new record instance with attributes set.

        :param attributes: the dict of attribute values to assign to the new record.
        :return: None
        :rtype: None
        """
        _new_record = self._record()
        _new_record.revision_id = self.last_id + 1
        _new_record.name = "New Revision"

        return _new_record



# -*- coding: utf-8 -*-
#
#       ramstk.models.dbrecord.baserecord.py is part of The RAMSTK Project
#
# All rights reserved.
# Copyright since 2007 Doyle "weibullguy" Rowland doyle.rowland <AT> reliaqual <DOT> com
"""Metaclasses for the database Record model."""

# RAMSTK Package Imports
from ramstk.utilities import none_to_default


class RAMSTKBaseRecord:
    """Metaclass for all RAMSTK Record models."""

    def set_attributes(self, attributes):
        """Set one or more RAMSTK<Table> attributes.

        .. note:: you should pop the primary and foreign key entries from the
            attributes dict before passing it to this method.

        :param attributes: dict of key:value pairs to assign to the
            instance attributes.
        :return: None
        :rtype: None
        :raise: AttributeError if passed an attribute key that doesn't exist as
            a table field.
        """
        for _key in attributes:
            getattr(self, _key)
            setattr(
                self,
                _key,
                none_to_default(attributes[_key], self.__defaults__[_key]),
            )



# Standard Library Imports
from typing import Tuple

# RAMSTK Package Imports
from ramstk.configuration import RAMSTKSiteConfiguration as RAMSTKSiteConfiguration
from ramstk.configuration import RAMSTKUserConfiguration as RAMSTKUserConfiguration
from ramstk.exim import Export as Export
from ramstk.exim import Import as Import
from ramstk.logger import RAMSTKLogManager as RAMSTKLogManager
from ramstk.models.db import BaseDatabase as BaseDatabase
from ramstk.models.db import RAMSTKProgramDB as RAMSTKProgramDB
from ramstk.models.dbtables import RAMSTKActionTable as RAMSTKActionTable
from ramstk.models.dbtables import RAMSTKAllocationTable as RAMSTKAllocationTable
from ramstk.models.dbtables import RAMSTKCauseTable as RAMSTKCauseTable
from ramstk.models.dbtables import RAMSTKControlTable as RAMSTKControlTable
from ramstk.models.dbtables import (
    RAMSTKDesignElectricTable as RAMSTKDesignElectricTable,
)
from ramstk.models.dbtables import (
    RAMSTKDesignMechanicTable as RAMSTKDesignMechanicTable,
)
from ramstk.models.dbtables import RAMSTKEnvironmentTable as RAMSTKEnvironmentTable
from ramstk.models.dbtables import (
    RAMSTKFailureDefinitionTable as RAMSTKFailureDefinitionTable,
)
from ramstk.models.dbtables import RAMSTKFunctionTable as RAMSTKFunctionTable
from ramstk.models.dbtables import RAMSTKHardwareTable as RAMSTKHardwareTable
from ramstk.models.dbtables import RAMSTKHazardTable as RAMSTKHazardTable
from ramstk.models.dbtables import RAMSTKMechanismTable as RAMSTKMechanismTable
from ramstk.models.dbtables import RAMSTKMILHDBK217FTable as RAMSTKMILHDBK217FTable
from ramstk.models.dbtables import RAMSTKMissionPhaseTable as RAMSTKMissionPhaseTable
from ramstk.models.dbtables import RAMSTKMissionTable as RAMSTKMissionTable
from ramstk.models.dbtables import RAMSTKModeTable as RAMSTKModeTable
from ramstk.models.dbtables import RAMSTKNSWCTable as RAMSTKNSWCTable
from ramstk.models.dbtables import RAMSTKOpLoadTable as RAMSTKOpLoadTable
from ramstk.models.dbtables import RAMSTKOpStressTable as RAMSTKOpStressTable
from ramstk.models.dbtables import RAMSTKProgramInfoTable as RAMSTKProgramInfoTable
from ramstk.models.dbtables import RAMSTKProgramStatusTable as RAMSTKProgramStatusTable
from ramstk.models.dbtables import RAMSTKReliabilityTable as RAMSTKReliabilityTable
from ramstk.models.dbtables import RAMSTKRequirementTable as RAMSTKRequirementTable
from ramstk.models.dbtables import RAMSTKRevisionTable as RAMSTKRevisionTable
from ramstk.models.dbtables import RAMSTKSimilarItemTable as RAMSTKSimilarItemTable
from ramstk.models.dbtables import RAMSTKSiteInfoTable as RAMSTKSiteInfoTable
from ramstk.models.dbtables import RAMSTKStakeholderTable as RAMSTKStakeholderTable
from ramstk.models.dbtables import RAMSTKTestMethodTable as RAMSTKTestMethodTable
from ramstk.models.dbtables import RAMSTKValidationTable as RAMSTKValidationTable
from ramstk.models.dbviews import RAMSTKFMEAView as RAMSTKFMEAView
from ramstk.models.dbviews import RAMSTKHardwareBoMView as RAMSTKHardwareBoMView
from ramstk.models.dbviews import RAMSTKPoFView as RAMSTKPoFView
from ramstk.models.dbviews import RAMSTKUsageProfileView as RAMSTKUsageProfileView
from ramstk.utilities import file_exists as file_exists
from ramstk.views.gtk3 import Gtk as Gtk
from ramstk.views.gtk3 import RAMSTKDesktop as RAMSTKDesktop
from ramstk.views.gtk3 import _ as _
from ramstk.views.gtk3.widgets import RAMSTKDatabaseSelect as RAMSTKDatabaseSelect

def do_connect_to_site_db(conn_info) -> BaseDatabase: ...
def do_first_run(configuration: RAMSTKSiteConfiguration) -> None: ...
def do_initialize_databases(
    configuration: RAMSTKUserConfiguration, site_db: BaseDatabase
) -> RAMSTKProgramDB: ...
def do_initialize_loggers(log_file: str, log_level: str) -> RAMSTKLogManager: ...
def do_read_site_configuration() -> RAMSTKSiteConfiguration: ...
def do_read_user_configuration() -> Tuple[
    RAMSTKUserConfiguration, RAMSTKLogManager
]: ...
def the_one_ring() -> None: ...



# -*- coding: utf-8 -*-
#
#       ramstk.views.gtk3.pof.panel.py is part of the RAMSTK Project
#
# All rights reserved.
# Copyright since 2007 Doyle "weibullguy" Rowland doyle.rowland <AT> reliaqual <DOT> com
"""GTK3 PoF Panels."""

# Standard Library Imports
from typing import Any, Dict, List, Union

# Third Party Imports
import treelib
from pubsub import pub

# RAMSTK Package Imports
from ramstk.views.gtk3 import GdkPixbuf, Gtk, _
from ramstk.views.gtk3.widgets import RAMSTKTreePanel


class PoFTreePanel(RAMSTKTreePanel):
    """Panel to display Physics if Failure analysis worksheet."""

    # Define private dictionary class attributes.
    _dic_visible_mask: Dict[str, List[bool]] = {
        "mode": [
            True,
            False,
            False,
            False,
            False,
            False,
            True,
            True,
            True,
            True,
            False,
            False,
            False,
            False,
            False,
            True,
        ],
        "mechanism": [
            True,
            False,
            False,
            False,
            False,
            False,
            True,
            False,
            False,
            False,
            False,
            False,
            False,
            False,
            False,
            False,
        ],
        "opload": [
            True,
            False,
            False,
            False,
            False,
            False,
            True,
            False,
            False,
            False,
            True,
            False,
            False,
            False,
            True,
            False,
        ],
        "opstress": [
            True,
            False,
            False,
            False,
            False,
            False,
            True,
            False,
            False,
            False,
            False,
            True,
            True,
            False,
            False,
            True,
        ],
        "test_method": [
            True,
            False,
            False,
            False,
            False,
            False,
            True,
            False,
            False,
            False,
            False,
            False,
            False,
            True,
            False,
            True,
        ],
    }

    # Define private list class attributes.

    # Define private scalar class attributes.
    _select_msg = "succeed_retrieve_pof"
    _tag = "pof"
    _title = _("Physics of Failure (PoF) Analysis")

    # Define public dictionary class attributes.

    # Define public list class attributes.

    # Define public scalar class attributes.

    def __init__(self) -> None:
        """Initialize an instance of the PoF analysis worksheet."""
        super().__init__()

        # Initialize private dictionary instance attributes.
        self.tvwTreeView.dic_row_loader = {
            "mode": self.__do_load_mode,
            "mechanism": self.__do_load_mechanism,
            "opload": self.__do_load_opload,
            "opstress": self.__do_load_opstress,
            "test_method": self.__do_load_test_method,
        }

        # Initialize private list instance attributes.

        # Initialize private scalar instance attributes.
        self._filtered_tree = True
        self._on_edit_message: str = f"wvw_editing_{self._tag}"

        # Initialize public dictionary instance attributes.
        self.dic_attribute_widget_map: Dict[str, List[Any]] = {
            "hardware_id": [
                0,
                Gtk.CellRendererText(),
                "edited",
                None,
                self._on_edit_message,
                0,
                {
                    "bg_color": "#FFFFFF",
                    "editable": False,
                    "fg_color": "#000000",
                    "visible": False,
                },
                _("Hardware ID"),
                "gint",
            ],
            "mode_id": [
                1,
                Gtk.CellRendererText(),
                "edited",
                None,
                self._on_edit_message,
                0,
                {
                    "bg_color": "#FFFFFF",
                    "editable": False,
                    "fg_color": "#000000",
                    "visible": False,
                },
                _("Mode ID"),
                "gint",
            ],
            "mechanism_id": [
                2,
                Gtk.CellRendererText(),
                "edited",
                None,
                self._on_edit_message,
                0,
                {
                    "bg_color": "#FFFFFF",
                    "editable": False,
                    "fg_color": "#000000",
                    "visible": False,
                },
                _("Mechanism ID"),
                "gint",
            ],
            "opload_id": [
                3,
                Gtk.CellRendererText(),
                "edited",
                None,
                self._on_edit_message,
                0,
                {
                    "bg_color": "#FFFFFF",
                    "editable": False,
                    "fg_color": "#000000",
                    "visible": False,
                },
                _("Load ID"),
                "gint",
            ],
            "opstress_id": [
                4,
                Gtk.CellRendererText(),
                "edited",
                None,
                self._on_edit_message,
                0,
                {
                    "bg_color": "#FFFFFF",
                    "editable": False,
                    "fg_color": "#000000",
                    "visible": False,
                },
                _("Stress ID"),
                "gint",
            ],
            "test_method_id": [
                5,
                Gtk.CellRendererText(),
                "edited",
                None,
                self._on_edit_message,
                0,
                {
                    "bg_color": "#FFFFFF",
                    "editable": False,
                    "fg_color": "#000000",
                    "visible": False,
                },
                _("Test ID"),
                "gint",
            ],
            "description": [
                6,
                Gtk.CellRendererText(),
                "edited",
                self._on_cell_edit,
                "wvw_editing_pof",
                "",
                {
                    "bg_color": "#FFFFFF",
                    "editable": True,
                    "fg_color": "#000000",
                    "visible": True,
                },
                _("Description"),
                "gchararray",
            ],
            "effect_end": [
                7,
                Gtk.CellRendererText(),
                "edited",
                super().on_cell_edit,
                self._on_edit_message,
                "",
                {
                    "bg_color": "#FFFFFF",
                    "editable": True,
                    "fg_color": "#000000",
                    "visible": True,
                },
                _("End Effect"),
                "gchararray",
            ],
            "severity_class": [
                8,
                Gtk.CellRendererText(),
                "edited",
                super().on_cell_edit,
                self._on_edit_message,
                "",
                {
                    "bg_color": "#FFFFFF",
                    "editable": True,
                    "fg_color": "#000000",
                    "visible": True,
                },
                _("Severity"),
                "gchararray",
            ],
            "mode_probability": [
                9,
                Gtk.CellRendererText(),
                "edited",
                super().on_cell_edit,
                self._on_edit_message,
                0.0,
                {
                    "bg_color": "#FFFFFF",
                    "editable": True,
                    "fg_color": "#000000",
                    "visible": True,
                },
                _("Mode Probability"),
                "gfloat",
            ],
            "damage_model": [
                10,
                Gtk.CellRendererCombo(),
                "edited",
                super().on_cell_edit,
                "wvw_editing_opload",
                "",
                {
                    "bg_color": "#FFFFFF",
                    "editable": True,
                    "fg_color": "#000000",
                    "visible": True,
                },
                _("Damage Model"),
                "gchararray",
            ],
            "measurable_parameter": [
                11,
                Gtk.CellRendererCombo(),
                "edited",
                super().on_cell_edit,
                "wvw_editing_opstress",
                "",
                {
                    "bg_color": "#FFFFFF",
                    "editable": True,
                    "fg_color": "#000000",
                    "visible": True,
                },
                _("Measurable Parameter"),
                "gchararray",
            ],
            "load_history": [
                12,
                Gtk.CellRendererCombo(),
                "edited",
                super().on_cell_edit,
                "wvw_editing_opstress",
                "",
                {
                    "bg_color": "#FFFFFF",
                    "editable": True,
                    "fg_color": "#000000",
                    "visible": True,
                },
                _("Load History Method"),
                "gchararray",
            ],
            "boundary_conditions": [
                13,
                Gtk.CellRendererText(),
                "edited",
                super().on_cell_edit,
                "wvw_editing_test_method",
                "",
                {
                    "bg_color": "#FFFFFF",
                    "editable": True,
                    "fg_color": "#000000",
                    "visible": True,
                },
                _("Boundary Conditions"),
                "gchararray",
            ],
            "priority_id": [
                14,
                Gtk.CellRendererSpin(),
                "edited",
                super().on_cell_edit,
                "wvw_editing_opload",
                0,
                {
                    "bg_color": "#FFFFFF",
                    "editable": True,
                    "fg_color": "#000000",
                    "visible": True,
                },
                _("Priority"),
                "gint",
            ],
            "remarks": [
                15,
                Gtk.CellRendererText(),
                "edited",
                self._on_cell_edit,
                "wvw_editing_pof",
                "",
                {
                    "bg_color": "#FFFFFF",
                    "editable": True,
                    "fg_color": "#000000",
                    "visible": True,
                },
                _("Remarks"),
                "gchararray",
            ],
        }
        self.dic_icons: Dict[str, str] = {}

        # Initialize public list instance attributes.
        self.lst_damage_models: List[str] = []
        self.lst_load_history: List[str] = []
        self.lst_measurable_parameters: List[str] = []

        # Initialize public scalar instance attributes.

        super().do_set_properties()
        super().do_make_panel()
        super().do_set_callbacks()

        self.tvwTreeView.set_tooltip_text(
            _(
                "Displays the Physics of Failure (PoF) Analysis for the currently "
                "selected hardware item."
            )
        )

        # Subscribe to PyPubSub messages.
        pub.subscribe(super().do_load_panel, "succeed_retrieve_pof")

        pub.subscribe(self._on_select_hardware, "selected_hardware")

    # pylint: disable=unused-argument
    # noinspection PyUnusedLocal
    def do_filter_tree(
        self, model: Gtk.TreeModel, row: Gtk.TreeIter, data: Any
    ) -> bool:
        """Filter PoF to show only those rows associated with the selected Hardware.

        :param model: the filtered model for the PoF RAMSTKTreeView.
        :param row: the iter to check against condition(s).
        :param data: unused in this method; required by Gtk.TreeModelFilter() widget.
        :return: True if row should be visible, False else.
        :rtype: bool
        """
        return model[row][0] == self._parent_id

    def do_get_pof_level(self, model: Gtk.TreeModel, row: Gtk.TreeIter) -> None:
        """Determine the FMEA level of the selected FMEA row.

        :param model: the FMEA Gtk.TreeModel().
        :param row: the selected Gtk.TreeIter() in the FMECA.
        :return: None
        :rtype: None
        """
        _cid = ""

        for _col in [1, 2, 3, 4, 5]:
            _cid = f"{_cid}{int(bool(model.get_value(row, _col)))}"

        self.level = {
            "10000": "mode",
            "11000": "mechanism",
            "11100": "opload",
            "11110": "opstress",
            "11101": "test_method",
        }[_cid]

    def do_load_comboboxes(self) -> None:
        """Load the RAMSTKComboBox() widgets.

        :return: None
        :rtype: None
        """
        self.__do_load_damage_models()
        self.__do_load_measureable_parameters()
        self.__do_load_load_history()

        # Set the priority Gtk.CellRendererSpin()'s adjustment limits and
        # step increments.
        _cell = self.tvwTreeView.get_column(
            self.tvwTreeView.position["priority_id"]
        ).get_cells()[0]
        _adjustment = _cell.get_property("adjustment")
        _adjustment.configure(5, 1, 5, -1, 0, 0)

    def _on_cell_edit(
        self,
        cell: Gtk.CellRenderer,
        path: str,
        new_text: str,
        key: str,
        message: str,
    ) -> None:
        """Handle edits of description column to ensure proper level is updated.

        :param cell: the Gtk.CellRenderer() that was edited.
        :param path: the RAMSTKTreeView() path of the Gtk.CellRenderer()
            that was edited.
        :param new_text: the new text in the edited Gtk.CellRenderer().
        :param key: the column key of the edited Gtk.CellRenderer().
        :param message: the PyPubSub message to publish.
        :return: None
        """
        super().on_cell_edit(
            cell,
            path,
            new_text,
            key,
            f"wvw_editing_{self.level}",
        )

    def _on_row_change(self, selection: Gtk.TreeSelection) -> None:
        """Handle events for the PoF Work View RAMSTKTreeView().

        This method is called whenever a RAMSTKTreeView() row is activated.

        :param selection: the TreeSelection() of the currently
            selected row in the PoF RAMSTKTreeView().
        :return: None
        """
        _attributes = super().on_row_change(selection)
        _model, _row = selection.get_selected()

        if _row is not None:
            self.do_get_pof_level(_model, _row)
            super().do_set_visible_columns(_attributes)
            self._record_id = _attributes[f"{self.level}_id"]

    def _on_select_hardware(
        self, attributes: Dict[str, Union[int, float, str]]
    ) -> None:
        """Filter FMEA when Hardware is selected.

        :param attributes: the dict of attributes for the selected Hardware.
        :return: None
        :rtype: None
        """
        self._parent_id = attributes["hardware_id"]
        self.tvwTreeView.filt_model.refilter()

    def __do_load_damage_models(self) -> None:
        """Load the RAMSTKTreeView() damage model CellRendererCombo().

        :return: None
        """
        self.tvwTreeView.do_load_combo_cell(
            self.tvwTreeView.position["damage_model"], self.lst_damage_models
        )

    def __do_load_load_history(self) -> None:
        """Load the operating load history CellRendererCombo().

        :return: None
        :rtype: None
        """
        self.tvwTreeView.do_load_combo_cell(
            self.tvwTreeView.position["load_history"], self.lst_load_history
        )

    def __do_load_measureable_parameters(self) -> None:
        """Load the measureable parameters CellRendererCombo().

        :return: None
        """
        self.tvwTreeView.do_load_combo_cell(
            self.tvwTreeView.position["measurable_parameter"],
            self.lst_measurable_parameters,
        )

    def __do_load_mechanism(
        self, node: treelib.Node, row: Gtk.TreeIter
    ) -> Gtk.TreeIter:
        """Load a failure mechanism record into the RAMSTKTreeView().

        :param node: the treelib Node() with the mechanism data to load.
        :param row: the parent row of the mechanism to load into the FMEA form.
        :return: _new_row; the row that was just populated with mechanism data.
        """
        _new_row = None

        [[__, _entity]] = node.data.items()  # pylint: disable=unused-variable

        # noinspection PyArgumentList
        _icon = GdkPixbuf.Pixbuf.new_from_file_at_size(
            self.dic_icons["mechanism"], 22, 22
        )

        _attributes = [
            _entity.hardware_id,
            _entity.mode_id,
            _entity.mechanism_id,
            0,
            0,
            0,
            _entity.description,
            "",
            "",
            0.0,
            "",
            "",
            "",
            "",
            0,
            "",
            _icon,
        ]

        try:
            _new_row = self.tvwTreeView.unfilt_model.append(row, _attributes)
        except (AttributeError, TypeError, ValueError):
            _new_row = None
            _message = _(
                f"An error occurred when loading failure mechanism {node.identifier} "
                f"in the physics of failure analysis.  This might indicate it was "
                f"missing it's data package, some of the data in the package was "
                f"missing, or some of the data was the wrong type.  Row data "
                f"was: {_attributes}"
            )
            pub.sendMessage(
                "do_log_warning_msg", logger_name="WARNING", message=_message
            )

        return _new_row

    def __do_load_mode(self, node: treelib.Node, row: Gtk.TreeIter) -> Gtk.TreeIter:
        """Load a failure mode record into the RAMSTKTreeView().

        :param node: the treelib Node() with the mode data to load.
        :param row: the parent row of the mode to load into the FMEA form.
        :return: _new_row; the row that was just populated with mode data.
        """
        _new_row = None

        [[__, _entity]] = node.data.items()  # pylint: disable=unused-variable

        # noinspection PyArgumentList
        _icon = GdkPixbuf.Pixbuf.new_from_file_at_size(self.dic_icons["mode"], 22, 22)

        _attributes = [
            _entity.hardware_id,
            _entity.mode_id,
            0,
            0,
            0,
            0,
            _entity.description,
            _entity.effect_end,
            _entity.severity_class,
            _entity.mode_ratio,
            "",
            "",
            "",
            "",
            0,
            "",
            _icon,
        ]

        try:
            _new_row = self.tvwTreeView.unfilt_model.append(row, _attributes)
        except (AttributeError, TypeError, ValueError):
            _new_row = None
            _message = _(
                f"An error occurred when loading failure mode {node.identifier} in the "
                f"physics of failure analysis.  This might indicate it was missing "
                f"its data package, some of the data in the package was missing, or "
                f"some of the data was the wrong type.  Row data was: {_attributes}"
            )
            pub.sendMessage(
                "do_log_warning_msg", logger_name="WARNING", message=_message
            )

        return _new_row

    def __do_load_opload(self, node: treelib.Node, row: Gtk.TreeIter) -> Gtk.TreeIter:
        """Load a failure mechanism record into the RAMSTKTreeView().

        :param node: the treelib Node() with the mechanism data to load.
        :param row: the parent row of the mechanism to load into the FMEA form.
        :return: _new_row; the row that was just populated with mechanism data.
        """
        _new_row = None

        [[__, _entity]] = node.data.items()  # pylint: disable=unused-variable

        # noinspection PyArgumentList
        _icon = GdkPixbuf.Pixbuf.new_from_file_at_size(self.dic_icons["opload"], 22, 22)

        _attributes = [
            _entity.hardware_id,
            _entity.mode_id,
            _entity.mechanism_id,
            _entity.opload_id,
            0,
            0,
            _entity.description,
            "",
            "",
            0.0,
            _entity.damage_model,
            "",
            "",
            "",
            _entity.priority_id,
            "",
            _icon,
        ]

        try:
            _new_row = self.tvwTreeView.unfilt_model.append(row, _attributes)
        except (AttributeError, TypeError, ValueError):
            _new_row = None
            _message = _(
                f"An error occurred when loading operating load {node.identifier} in "
                f"the physics of failure analysis.  This might indicate it was "
                f"missing its data package, some of the data in the package was "
                f"missing, or some of the data was the wrong type.  Row data "
                f"was: {_attributes}"
            )
            pub.sendMessage(
                "do_log_warning_msg", logger_name="WARNING", message=_message
            )

        return _new_row

    def __do_load_opstress(self, node: treelib.Node, row: Gtk.TreeIter) -> Gtk.TreeIter:
        """Load a failure mechanism record into the RAMSTKTreeView().

        :param node: the treelib Node() with the mechanism data to load.
        :param row: the parent row of the mechanism to load into the FMEA form.
        :return: _new_row; the row that was just populated with mechanism data.
        """
        _new_row = None

        [[__, _entity]] = node.data.items()  # pylint: disable=unused-variable

        # noinspection PyArgumentList
        _icon = GdkPixbuf.Pixbuf.new_from_file_at_size(
            self.dic_icons["opstress"], 22, 22
        )

        _attributes = [
            _entity.hardware_id,
            _entity.mode_id,
            _entity.mechanism_id,
            _entity.opload_id,
            _entity.opstress_id,
            0,
            _entity.description,
            "",
            "",
            0.0,
            "",
            _entity.measurable_parameter,
            _entity.load_history,
            "",
            0,
            _entity.remarks,
            _icon,
        ]

        try:
            _new_row = self.tvwTreeView.unfilt_model.append(row, _attributes)
        except (AttributeError, TypeError, ValueError):
            _new_row = None
            _message = _(
                f"An error occurred when loading operating stress {node.identifier} in "
                f"the physics of failure analysis.  This might indicate it was "
                f"missing its data package, some of the data in the package was "
                f"missing, or some of the data was the wrong type.  Row data "
                f"was: {_attributes}"
            )
            pub.sendMessage(
                "do_log_warning_msg", logger_name="WARNING", message=_message
            )

        return _new_row

    def __do_load_test_method(
        self, node: treelib.Node, row: Gtk.TreeIter
    ) -> Gtk.TreeIter:
        """Load a failure mechanism record into the RAMSTKTreeView().

        :param node: the treelib Node() with the mechanism data to load.
        :param row: the parent row of the mechanism to load into the FMEA form.
        :return: _new_row; the row that was just populated with mechanism data.
        """
        _new_row = None

        [[__, _entity]] = node.data.items()  # pylint: disable=unused-variable

        # noinspection PyArgumentList
        _icon = GdkPixbuf.Pixbuf.new_from_file_at_size(
            self.dic_icons["test_method"], 22, 22
        )

        _attributes = [
            _entity.hardware_id,
            _entity.mode_id,
            _entity.mechanism_id,
            _entity.opload_id,
            0,
            _entity.test_method_id,
            _entity.description,
            "",
            "",
            0.0,
            "",
            "",
            "",
            _entity.boundary_conditions,
            0,
            _entity.remarks,
            _icon,
        ]

        try:
            _new_row = self.tvwTreeView.unfilt_model.append(row, _attributes)
        except (AttributeError, TypeError, ValueError):
            _new_row = None
            _message = _(
                f"An error occurred when loading test method {node.identifier} in the "
                f"physics of failure analysis.  This might indicate it was missing its "
                f"data package, some of the data in the package was missing, or some "
                f"of the data was the wrong type.  Row data was: {_attributes}"
            )
            pub.sendMessage(
                "do_log_warning_msg", logger_name="WARNING", message=_message
            )

        return _new_row



# Standard Library Imports
from typing import Dict

# RAMSTK Package Imports
from ramstk.configuration import RAMSTKUserConfiguration as RAMSTKUserConfiguration
from ramstk.db import BaseDatabase as BaseDatabase
from ramstk.db import do_create_program_db as do_create_program_db
from ramstk.exceptions import DataAccessError as DataAccessError

class RAMSTKProgramDB:
    dic_tables: Dict[str, object]
    dic_views: Dict[str, object]
    user_configuration: RAMSTKUserConfiguration
    program_dao: BaseDatabase
    def __init__(self) -> None: ...
    def do_create_program(
        self, program_db: BaseDatabase, database: Dict[str, str]
    ) -> None: ...
    def do_open_program(
        self, program_db: BaseDatabase, database: Dict[str, str]
    ) -> None: ...
    def do_close_program(self) -> None: ...
    @staticmethod
    def do_save_program() -> None: ...



# -*- coding: utf-8 -*-
#
#       ramstk.models.dbtables.programdb_allocation_table.py is part of The RAMSTK
#       Project
#
# All rights reserved.
# Copyright since 2007 Doyle "weibullguy" Rowland doyle.rowland <AT> reliaqual <DOT> com
"""RAMSTKAllocation Table Model."""

# Standard Library Imports
from datetime import date
from typing import Any, Dict, Tuple, Type, Union

# Third Party Imports
import treelib
from pubsub import pub

# RAMSTK Package Imports
from ramstk.analyses import allocation

# RAMSTK Local Imports
from ..dbrecords import RAMSTKAllocationRecord
from .basetable import RAMSTKBaseTable


class RAMSTKAllocationTable(RAMSTKBaseTable):
    """Contain the attributes and methods of the Allocation table model.

    :cvar _system_hazard_rate: the hazard rate of the system under analysis.

    :ivar _node_hazard_rate: the hazard rate of the selected node.
    """

    # Define private dictionary class attributes.

    # Define private list class attributes.

    # Define private scalar class attributes.
    _db_id_colname = "fld_hardware_id"
    _db_tablename = "ramstk_allocation"
    _select_msg = "selected_revision"
    _tag = "allocation"

    _system_hazard_rate: float = 0.0

    # Define public dictionary class attributes.

    # Define public list class attributes.

    # Define public scalar class attributes.

    def __init__(self, **kwargs: Dict[str, Union[float, int, str]]) -> None:
        """Initialize a RAMSTKAllocation table model instance."""
        super().__init__(**kwargs)

        # Initialize private dictionary attributes.

        # Initialize private list attributes.
        self._lst_id_columns = [
            "revision_id",
            "hardware_id",
            "record_id",
        ]

        # Initialize private scalar attributes.
        self._node_hazard_rate: float = 0.0
        self._record: Type[RAMSTKAllocationRecord] = RAMSTKAllocationRecord

        # Initialize public dictionary attributes.

        # Initialize public list attributes.

        # Initialize public scalar attributes.
        self.pkey = "hardware_id"

        # Subscribe to PyPubSub messages.
        pub.subscribe(
            self.do_calculate_agree_allocation,
            "request_calculate_agree_allocation",
        )
        pub.subscribe(
            self.do_calculate_arinc_allocation,
            "request_calculate_arinc_allocation",
        )
        pub.subscribe(
            self.do_calculate_equal_allocation,
            "request_calculate_equal_allocation",
        )
        pub.subscribe(
            self.do_calculate_foo_allocation,
            "request_calculate_foo_allocation",
        )
        pub.subscribe(
            self.do_calculate_allocation_goals,
            "request_calculate_allocation_goals",
        )
        pub.subscribe(
            self._on_insert_hardware,
            "succeed_insert_hardware",
        )

    def do_get_new_record(  # pylint: disable=method-hidden
        self, attributes: Dict[str, Union[date, float, int, str]]
    ) -> RAMSTKAllocationRecord:
        """Gets a new record instance with attributes set.

        :param attributes: the dict of attribute values to assign to the new record.
        :return: None
        :rtype: None
        """
        # pylint: disable=attribute-defined-outside-init
        self._parent_id = attributes["parent_id"]  # type: ignore

        _new_record = self._record()
        _new_record.revision_id = attributes["revision_id"]
        _new_record.parent_id = attributes["parent_id"]
        _new_record.hardware_id = attributes["hardware_id"]

        return _new_record

    def do_calculate_allocation_goals(self, node_id: int) -> None:
        """Calculate the allocation goals.

        :param node_id: the record ID whose goals are to be calculated.
        :return: None
        :rtype: None
        """
        _record = self.tree.get_node(node_id).data[self._tag]
        _attributes = _record.get_attributes()

        _attributes = allocation.do_calculate_goals(**_attributes)

        self.do_set_attributes(
            node_id=node_id,
            package={"hazard_rate_goal": _attributes["hazard_rate_goal"]},
        )
        self.do_set_attributes(
            node_id=node_id,
            package={"mtbf_goal": _attributes["mtbf_goal"]},
        )
        self.do_set_attributes(
            node_id=node_id,
            package={"reliability_goal": _attributes["reliability_goal"]},
        )

    def do_calculate_agree_allocation(self, node_id: int, duty_cycle: float) -> None:
        """Allocate reliability using the AGREE method.

        :param node_id: the record ID whose allocation is to be calculated.
        :param duty_cycle: the duty cycle of the item to be calculated.
        :return: None
        :rtype: None
        """
        _record = self.tree.get_node(node_id).data[self._tag]
        _parent_goal = _record.reliability_goal
        _method_id = _record.allocation_method_id

        for _node in self.tree.children(node_id):
            _attributes: Dict[str, Any] = _node.data["allocation"].get_attributes()
            _attributes["allocation_method_id"] = _method_id
            _attributes["duty_cycle"] = duty_cycle

            (
                _attributes["n_sub_elements"],
                _attributes["n_sub_systems"],
            ) = self._do_calculate_agree_total_elements(node_id)

            _attributes = allocation.do_allocate_reliability(
                _parent_goal, 0, **_attributes
            )

            for _attr in [
                "mtbf_alloc",
                "hazard_rate_alloc",
                "reliability_alloc",
                "reliability_goal",
            ]:
                self.do_set_attributes(
                    node_id=_node.identifier,
                    package={_attr: _attributes[_attr]},
                )

            _parent_goal = _node.data["allocation"].reliability_goal

        pub.sendMessage(
            "succeed_calculate_allocation",
            tree=self.tree,
        )

    def do_calculate_arinc_allocation(self, node_id: int) -> None:
        """Allocate reliability using the ARINC method.

        :param node_id: the record ID whose allocation is to be calculated.
        :return: None
        :rtype: None
        :raises: ZeroDivisionError if the system hazard rate is zero.
        """
        _record = self.tree.get_node(node_id).data[self._tag]

        _parent_goal = _record.hazard_rate_goal
        _method_id = _record.allocation_method_id

        for _node in self.tree.children(node_id):
            _attributes = _node.data["allocation"].get_attributes()
            _attributes["allocation_method_id"] = _method_id

            _attributes["weight_factor"] = (
                self._node_hazard_rate / self._system_hazard_rate
            )
            _attributes = allocation.do_allocate_reliability(
                _parent_goal, 0, **_attributes
            )

            for _attr in [
                "mtbf_alloc",
                "hazard_rate_alloc",
                "reliability_alloc",
                "reliability_goal",
            ]:
                self.do_set_attributes(
                    node_id=_node.identifier,
                    package={_attr: _attributes[_attr]},
                )

            _parent_goal = _node.data["allocation"].hazard_rate_goal

        pub.sendMessage(
            "succeed_calculate_allocation",
            tree=self.tree,
        )

    def do_calculate_equal_allocation(self, node_id: int) -> None:
        """Allocate reliability using equal apportionment.

        :param node_id: the record ID whose allocation is to be calculated.
        :return: None
        :rtype: None
        """
        _record = self.tree.get_node(node_id).data[self._tag]
        _attributes = _record.get_attributes()

        _parent_goal = _attributes["reliability_goal"]
        _method_id = _attributes["allocation_method_id"]

        for _node in self.tree.children(node_id):
            _attributes = _node.data["allocation"].get_attributes()
            _attributes["allocation_method_id"] = _method_id

            _attributes["weight_factor"] = 1.0 / _node.data["allocation"].n_sub_systems
            _attributes = allocation.do_allocate_reliability(
                _parent_goal, 0, **_attributes
            )

            for _attr in [
                "allocation_method_id",
                "weight_factor",
                "mtbf_alloc",
                "hazard_rate_alloc",
                "reliability_alloc",
                "reliability_goal",
            ]:
                self.do_set_attributes(
                    node_id=_node.identifier,
                    package={_attr: _attributes[_attr]},
                )

            _parent_goal = _attributes["reliability_alloc"]

        pub.sendMessage(
            "succeed_calculate_allocation",
            tree=self.tree,
        )

    def do_calculate_foo_allocation(self, node_id: int) -> None:
        """Allocate reliability using the FOO method.

        :param node_id: the record ID whose allocation is to be calculated.
        :return: None
        :rtype: None
        """
        _record = self.tree.get_node(node_id).data[self._tag]

        _parent_goal = _record.hazard_rate_goal
        _method_id = _record.allocation_method_id
        _cum_weight = self._do_calculate_foo_cumulative_weight(node_id)

        for _node in self.tree.children(node_id):
            _attributes = _node.data["allocation"].get_attributes()
            _attributes["allocation_method_id"] = _method_id

            _attributes = allocation.do_allocate_reliability(
                _parent_goal, _cum_weight, **_attributes
            )

            for _attr in [
                "weight_factor",
                "percent_weight_factor",
                "mtbf_alloc",
                "hazard_rate_alloc",
                "reliability_alloc",
                "reliability_goal",
            ]:
                self.do_set_attributes(
                    node_id=_node.identifier,
                    package={_attr: _attributes[_attr]},
                )

        pub.sendMessage(
            "succeed_calculate_allocation",
            tree=self.tree,
        )

    def _do_calculate_agree_total_elements(self, node_id: int) -> Tuple[int, int]:
        """Calculate the total number of elements for the AGREE method.

        :param node_id: the record ID whose allocation is to be calculated.
        :return: _n_sub_elements, _n_sub_systems; the number of
            sub-elements and subsystems comprising the selected node.
        :rtype: tuple
        """
        _n_sub_elements = 0
        _n_sub_systems = 0

        for _node in self.tree.children(node_id):
            _n_sub_elements += _node.data["allocation"].n_sub_elements
            _n_sub_systems += _node.data["allocation"].n_sub_systems

        return _n_sub_elements, _n_sub_systems

    def _do_calculate_foo_cumulative_weight(self, node_id: int) -> int:
        """Calculate the cumulative weight for the FOO method.

        :param node_id: the node (allocation) ID of the allocation item whose
            goal is to be allocated.
        :return: _cum_weight; the cumulative weighting factor for the
            allocation item to be allocated.
        """
        _cum_weight = 0

        for _node in self.tree.children(node_id):
            _attributes = _node.data["allocation"].get_attributes()
            _attributes["weight_factor"] = (
                _attributes["int_factor"]
                * _attributes["soa_factor"]
                * _attributes["op_time_factor"]
                * _attributes["env_factor"]
            )
            _cum_weight += _attributes["weight_factor"]

            self.do_set_attributes(
                node_id=_node.identifier,
                package={"weight_factor": _attributes["weight_factor"]},
            )

        return _cum_weight

    def _on_insert_hardware(self, tree: treelib.Tree) -> None:
        """Add new node to the Allocation tree for the newly added Hardware.

        Allocation records are added by triggers in the database when a new Hardware
        item is added.  This method simply adds a new node to the Allocation tree
        with a blank record.

        :param tree: the Hardware tree with the new node.
        :return: None
        :rtype: None
        """
        for _node in tree.all_nodes()[1:]:
            if (
                not self.tree.contains(_node.identifier)
                and not _node.data["hardware"].part
            ):
                _attributes = {
                    "revision_id": _node.data["hardware"].revision_id,
                    "hardware_id": _node.data["hardware"].hardware_id,
                    "parent_id": _node.data["hardware"].parent_id,
                }
                _record = self.do_get_new_record(_attributes)
                self.tree.create_node(
                    tag=self._tag,
                    identifier=_node.data["hardware"].hardware_id,
                    parent=_node.data["hardware"].parent_id,
                    data={self._tag: _record},
                )



# pylint: disable=unused-import
# -*- coding: utf-8 -*-
#
#       ramstk.views.gtk3.books.__init__.py is part of the RAMSTK Project
#
# All rights reserved.
# Copyright 2019 Doyle Rowland doyle.rowland <AT> reliaqual <DOT> com
"""RAMSTK GTK3 Books Package."""

# RAMSTK Local Imports
from .modulebook import RAMSTKModuleBook
from .workbook import RAMSTKWorkBook



# Standard Library Imports
from typing import Callable, Dict, List, Union

# Third Party Imports
import treelib

# RAMSTK Local Imports
from .baseview import RAMSTKBaseView as RAMSTKBaseView

class RAMSTKUsageProfileView(RAMSTKBaseView):
    _tag: str
    _dic_load_functions: Dict[str, Callable[..., object]]
    _dic_trees: Dict[str, treelib.Tree]
    _lst_modules: List[str]
    def __init__(self, **kwargs: Dict[str, Union[float, int, str]]) -> None: ...
    def _do_load_environments(self, mission_phase_id: int, parent_id: str) -> None: ...
    def _do_load_missions(self) -> None: ...
    def _do_load_mission_phases(self, mission_id: int) -> None: ...



# -*- coding: utf-8 -*-
#
#       ramstk.models.db.program_database.py is part of the RAMSTK Project
#
# All rights reserved.
# Copyright since 2007 Doyle "weibullguy" Rowland doyle.rowland <AT> reliaqual <DOT> com
"""The RAMSTK program database model."""

# Standard Library Imports
from typing import Dict

# Third Party Imports
from pubsub import pub

# RAMSTK Package Imports
from ramstk.configuration import RAMSTKUserConfiguration
from ramstk.exceptions import DataAccessError

# RAMSTK Local Imports
from .basedatabase import BaseDatabase, do_create_program_db


class RAMSTKProgramDB:
    """The RAMSTK program database model class.

    The attributes of a RAMSTK program database model are:

    :ivar dic_tables: a dict containing the instances of all the database table
        models this program database model is managing.
    :ivar dic_views: a dict containing the instances of all the database view
        models this program database model is managing.
    :ivar user_configuration: the RAMSTKUserConfiguration instance associated with
        this progam database model.
    :ivar program_dao: the BaseDatabase() object that will connect to the
        RAMSTK program database.
    """

    def __init__(self) -> None:
        """Initialize an instance of the RAMSTK program database model."""

        # Initialize private dictionary attributes.

        # Initialize private list attributes.

        # Initialize private scalar attributes.

        # Initialize public dictionary attributes.
        self.dic_tables: Dict[str, object] = {
            "action": object,
            "allocation": object,
            "cause": object,
            "control": object,
            "design_electric": object,
            "design_mechanic": object,
            "environment": object,
            "failure_definition": object,
            "function": object,
            "hardware": object,
            "hazards": object,
            "mechanism": object,
            "milhdbk217f": object,
            "mission": object,
            "mission_phase": object,
            "mode": object,
            "nswc": object,
            "opload": object,
            "opstress": object,
            "program_info": object,
            "program_status": object,
            "reliability": object,
            "requirement": object,
            "revision": object,
            "similar_item": object,
            "stakeholder": object,
            "test_method": object,
            "validation": object,
            "options": object,
            "export": object,
            "import": object,
        }
        self.dic_views: Dict[str, object] = {
            "fmea": object,
            "hardwarebom": object,
            "pof": object,
            "usage_profile": object,
        }

        # Initialize public list attributes.

        # Initialize public scalar attributes.
        self.user_configuration: RAMSTKUserConfiguration = RAMSTKUserConfiguration()
        self.program_dao: BaseDatabase = BaseDatabase()

        # Subscribe to PyPubSub messages.
        pub.subscribe(self.do_create_program, "request_create_program")
        pub.subscribe(self.do_open_program, "request_open_program")
        pub.subscribe(self.do_open_program, "succeed_create_program_database")
        pub.subscribe(self.do_close_program, "request_close_program")
        pub.subscribe(self.do_save_program, "request_update_program")

    def do_create_program(
        self, program_db: BaseDatabase, database: Dict[str, str]
    ) -> None:
        """Create a new RAMSTK Program database.

        :param program_db: the BaseDatabase() that is to be used to create and connect
            to the new RAMSTK program database.
        :param database: a dict containing the database connection arguments.
        :return: None
        :rtype: None
        """
        with open(
            f"{self.user_configuration.RAMSTK_CONF_DIR}/"
            f"{database['dialect']}_program_db.sql",
            "r",
            encoding="utf-8",
        ) as _sql_file:
            self.program_dao = program_db
            do_create_program_db(database, _sql_file)
            pub.sendMessage(
                "succeed_create_program_database",
                program_db=self.program_dao,
                database=database,
            )

    def do_open_program(
        self, program_db: BaseDatabase, database: Dict[str, str]
    ) -> None:
        """Open an RAMSTK Program database for analyses.

        :param program_db: the BaseDatabase() that is to be used to create and
            connect to the new RAMSTK program database.
        :param database: a dictionary of parameters to pass to the DAO.
        :return: None
        :rtype: None
        """
        self.program_dao = program_db

        try:
            self.program_dao.do_connect(database)
            pub.sendMessage("succeed_connect_program_database", dao=self.program_dao)
            pub.sendMessage(
                "request_retrieve_revisions", attributes={"revision_id": None}
            )
        except DataAccessError as _error:
            pub.sendMessage("fail_connect_program_database", error_message=_error.msg)

    def do_close_program(self) -> None:
        """Close the open RAMSTK Program database.

        :return: None
        :rtype: None
        :raises: AttributeError if a database isn't open.
        """
        try:
            self.program_dao.do_disconnect()
            pub.sendMessage("succeed_disconnect_program_database")
        except AttributeError:
            _error_msg = "Not currently connected to a database.  Nothing to close."
            pub.sendMessage(
                "fail_disconnect_program_database", error_message=_error_msg
            )

    @staticmethod
    def do_save_program() -> None:
        """Save the open RAMSTK Program database.

        :return: None
        :rtype: None
        """
        pub.sendMessage("request_update_all_revision")
        pub.sendMessage("request_update_all_function")
        pub.sendMessage("request_update_all_requirement")
        pub.sendMessage("request_update_all_stakeholder")
        pub.sendMessage("request_update_all_hardware")
        pub.sendMessage("request_update_all_validation")



# -*- coding: utf-8 -*-
#
#       ramstk.views.gtk3.usage_profile.view.py is part of The RAMSTK Project
#
# All rights reserved.
# Copyright since 2007 Doyle "weibullguy" Rowland doyle.rowland <AT> reliaqual <DOT> com
"""GTK3 Usage Profile Views."""

# Standard Library Imports
from typing import Dict

# Third Party Imports
from pubsub import pub

# RAMSTK Package Imports
from ramstk.configuration import RAMSTKUserConfiguration
from ramstk.logger import RAMSTKLogManager
from ramstk.views.gtk3 import Gtk, _
from ramstk.views.gtk3.widgets import RAMSTKWorkView

# RAMSTK Local Imports
from . import UsageProfileTreePanel


class UsageProfileWorkView(RAMSTKWorkView):
    """Display Usage Profiles associated with the selected Revision.

    The attributes of a Usage Profile List View are:

    :cvar _tag: the name of the module.

    :ivar _lst_mnu_labels: the list of labels for the view's pop-up
        menu.  The labels are listed in the order they appear in the menu.
    :ivar _lst_tooltips: the list of tooltips for the view's
        toolbar buttons and pop-up menu.  The tooltips are listed in the
        order they appear on the toolbar or pop-up menu.
    """

    # Define private dict class attributes.

    # Define private scalar class attributes.
    _tag: str = "usage_profile"
    _tablabel: str = "<span weight='bold'>" + _("Usage\nProfiles") + "</span>"
    _tabtooltip: str = _("Displays usage profiles for the selected revision.")

    # Define public dictionary class attributes.

    # Define public list class attributes.

    # Define public scalar class attributes.

    def __init__(
        self, configuration: RAMSTKUserConfiguration, logger: RAMSTKLogManager
    ) -> None:
        """Initialize an instance of the Usage Profile list view.

        :param configuration: the RAMSTK Configuration class instance.
        :param logger: the RAMSTKLogManager class instance.
        """
        super().__init__(configuration, logger)

        # Initialize private dictionary attributes.

        # Initialize private list attributes.
        self._lst_callbacks.insert(0, self._do_request_insert_sibling)
        self._lst_callbacks.insert(1, self._do_request_insert_child)
        self._lst_callbacks.insert(2, self._do_request_delete)
        self._lst_icons.insert(0, "insert_sibling")
        self._lst_icons.insert(1, "insert_child")
        self._lst_icons.insert(2, "remove")
        self._lst_mnu_labels = [
            _("Add Sibling"),
            _("Add Child"),
            _("Delete Selected"),
            _("Save Selected"),
            _("Save Profile"),
        ]
        self._lst_tooltips = [
            _(
                "Add a new usage profile entity at the same level "
                "as the currently selected entity."
            ),
            _(
                "Add a new usage profile entity one level below the "
                "currently selected entity."
            ),
            _("Delete the currently selected entity from the usage profile."),
            _("Save changes to the currently selected entity in the usage profile."),
            _("Save changes to all entities at the same level in the usage profile."),
        ]

        # Initialize private scalar attributes.
        self._pnlPanel = UsageProfileTreePanel()

        # Initialize public dictionary attributes.

        # Initialize public list attributes.

        # Initialize public scalar attributes.

        self.__make_ui()

        # Subscribe to PyPubSub messages.
        pub.subscribe(super().do_set_record_id, f"selected_{self._tag}")

    # pylint: disable=unused-argument
    def _do_request_delete(self, __button: Gtk.ToolButton) -> None:
        """Request to delete the selected Usage Profile record.

        :param __button: the Gtk.ToolButton() that called this method.
        :return: None
        """
        _parent = self.get_parent().get_parent().get_parent().get_parent().get_parent()
        _dialog = super().do_raise_dialog(parent=_parent)
        _dialog.do_set_message(
            message=_(
                f"You are about to delete {self._tag} {self.dic_pkeys['record_id']} "
                f"and all data associated with it.  Is this really what you want to do?"
            )
        )
        _dialog.do_set_message_type(message_type="question")

        if _dialog.do_run() == Gtk.ResponseType.YES:
            super().do_set_cursor_busy()
            pub.sendMessage(
                f"request_delete_{self._tag}",
                node_id=self.dic_pkeys["record_id"],
            )

        _dialog.do_destroy()

    # pylint: disable=unused-argument
    def _do_request_insert_child(self, __button: Gtk.ToolButton) -> None:
        """Request to add an entity to the Usage Profile.

        :return: None
        """
        super().do_set_cursor_busy()

        _attributes = self.__do_get_usage_profile_ids()

        if self._pnlPanel.level == "mission":
            _level = "mission_phase"
            _no_keys = ["environment_id"]
        elif self._pnlPanel.level == "mission_phase":
            _level = "environment"
            _no_keys = []
        else:
            _error = _("An environmental condition cannot have a child.")
            _parent = (
                self.get_parent().get_parent().get_parent().get_parent().get_parent()
            )
            _dialog = super().do_raise_dialog(parent=_parent)
            _dialog.do_set_message(message=_error)
            _dialog.do_set_message_type(message_type="error")
            _dialog.do_run()
            _dialog.do_destroy()
            pub.sendMessage(
                "fail_insert_usage_profile",
                error_message=_error,
            )
            return

        for _key in _no_keys:
            _attributes.pop(_key)

        super().do_set_cursor_busy()

        pub.sendMessage(f"request_insert_{_level}", attributes=_attributes)

    # pylint: disable=unused-argument
    def _do_request_insert_sibling(self, __button: Gtk.ToolButton) -> None:
        """Request to add a sibling entity to the Usage Profile.

        :return: None
        """
        super().do_set_cursor_busy()

        _attributes = self.__do_get_usage_profile_ids()

        if self._tag == "mission":
            _attributes.pop("mission_phase_id")
            _attributes.pop("environment_id")
        elif self._tag == "mission_phase":
            _attributes.pop("environment_id")

        pub.sendMessage(
            f"request_insert_{self._tag}",
            attributes=_attributes,
        )

    def __do_get_usage_profile_ids(self) -> Dict[str, int]:
        """Read each of the ID columns.

        :return: _attributes
        :rtype: dict
        """
        _attributes = {
            "revision_id": self._revision_id,
            "mission_id": 0,
            "mission_phase_id": 0,
            "environment_id": 0,
            "parent_id": 0,
            "record_id": 0,
        }

        (
            _model,
            _row,
        ) = self._pnlPanel.tvwTreeView.get_selection().get_selected()

        _attributes["mission_id"] = _model.get_value(_row, 1)
        _attributes["mission_phase_id"] = _model.get_value(_row, 2)
        _attributes["environment_id"] = _model.get_value(_row, 3)

        return _attributes

    def __make_ui(self) -> None:
        """Build the user interface for the usage profile list view.

        :return: None
        :rtype: None
        """
        super().do_make_layout()
        super().do_embed_treeview_panel()

        self._pnlPanel.dic_units = (
            self.RAMSTK_USER_CONFIGURATION.RAMSTK_MEASUREMENT_UNITS
        )
        self._pnlPanel.do_load_comboboxes()

        self._pnlPanel.tvwTreeView.dic_handler_id[
            "button-press"
        ] = self._pnlPanel.tvwTreeView.connect(
            "button_press_event", super().on_button_press
        )
        for _element in ["mission", "mission_phase", "environment"]:
            self._pnlPanel.dic_icons[_element] = self._dic_icons[_element]



# Standard Library Imports
from datetime import date
from typing import Any, Callable, Dict, List, Union

# Third Party Imports
import treelib

# RAMSTK Package Imports
from ramstk.db.base import BaseDatabase as BaseDatabase
from ramstk.exceptions import DataAccessError as DataAccessError
from ramstk.views.gtk3 import _ as _

def do_clear_tree(tree: treelib.Tree) -> treelib.Tree: ...

class RAMSTKBaseTable:
    _db_id_colname: str
    _db_tablename: str
    _deprecated: bool
    _root: int
    _select_msg: str
    _tag: str
    _lst_id_columns: List[str]
    _parent_id: int
    _record: object
    _revision_id: int
    dao: BaseDatabase
    last_id: int
    pkey: str
    tree: treelib.Tree
    do_get_new_record: Callable[[Dict[str, Union[date, float, int, str]]], object]
    def __init__(self, **kwargs: Dict[str, Union[float, int, str]]) -> None: ...
    def do_connect(self, dao: BaseDatabase) -> None: ...
    def do_create_all_codes(self, prefix: str) -> None: ...
    def do_delete(self, node_id: int) -> None: ...
    def do_get_attributes(self, node_id: int) -> None: ...
    def do_get_tree(self) -> None: ...
    def do_insert(
        self, attributes: Dict[str, Union[date, float, int, str]]
    ) -> None: ...
    def do_select(self, node_id: Any) -> Any: ...
    def do_select_all(
        self, attributes: Dict[str, Union[date, float, int, str]]
    ) -> None: ...
    def do_set_attributes(
        self, node_id: int, package: Dict[str, Union[date, float, int, str]]
    ) -> None: ...
    def do_set_attributes_all(
        self, attributes: Dict[str, Union[date, float, int, str]]
    ) -> None: ...
    def do_set_tree(self, tree: treelib.Tree) -> None: ...
    def do_update(self, node_id: int) -> None: ...
    def do_update_all(self) -> None: ...



# Standard Library Imports
from typing import Callable, Dict, List, Union

# Third Party Imports
import treelib

# RAMSTK Package Imports
from ramstk.db.base import BaseDatabase as BaseDatabase

class RAMSTKBaseView:
    _root: int
    _tag: str
    _dic_load_functions: Dict[str, Callable[..., object]]
    _dic_trees: Dict[str, treelib.Tree]
    _lst_modules: List[str]
    _revision_id: int
    dao: BaseDatabase
    tree: treelib.Tree
    def __init__(self, **kwargs: Dict[str, Union[float, int, str]]) -> None: ...
    def do_set_tree(self, tree: treelib.Tree) -> None: ...
    def on_select_all(self) -> None: ...



# Standard Library Imports
from typing import Callable, Dict, List, Union

# Third Party Imports
import treelib

# RAMSTK Local Imports
from .baseview import RAMSTKBaseView as RAMSTKBaseView

class RAMSTKHardwareBoMView(RAMSTKBaseView):
    _tag: str
    _dic_load_functions: Dict[str, Callable[..., object]]
    _dic_stress_limits: Dict[int, List[float]]
    _dic_trees: Dict[str, treelib.Tree]
    _lst_modules: List[str]
    _hr_multiplier: float
    def __init__(self, **kwargs: Dict[str, Union[float, int, str]]) -> None: ...
    def do_calculate_assembly_hazard_rates(self, node_id: int) -> None: ...
    def do_calculate_cost(self, node_id: int) -> None: ...
    def do_calculate_hardware(self, node_id: int) -> None: ...
    def do_calculate_hazard_rates(self, node_id: int) -> None: ...
    def do_calculate_part_count(self, node_id: int) -> None: ...
    def do_calculate_part_hazard_rates(self, node_id: int) -> None: ...
    def do_calculate_part_stress(self, node_id: int) -> None: ...
    def do_calculate_power_dissipation(self, node_id: int) -> float: ...
    def do_make_composite_ref_des(self, node_id: int = ...) -> None: ...
    def _do_load_hardware(self) -> None: ...
    def _do_load_design_electric(self) -> None: ...
    def _do_load_design_mechanic(self) -> None: ...
    def _do_load_milhdbk217f(self) -> None: ...
    def _do_load_nswc(self) -> None: ...
    def _do_load_reliability(self) -> None: ...



# -*- coding: utf-8 -*-
#
#       ramstk.views.gtk3.desktop.py is part of The RAMSTK Project
#
# All rights reserved.
# Copyright since 2007 Doyle "weibulguy" Rowland doyle.rowland <AT> reliaqual <DOT> com
"""The RAMSTK GTK3 desktop."""

# Standard Library Imports
import locale
from typing import Dict, List, TypeVar

# Third Party Imports
# noinspection PyPackageRequirements
from pubsub import pub

# RAMSTK Package Imports
from ramstk.configuration import RAMSTKSiteConfiguration, RAMSTKUserConfiguration
from ramstk.logger import RAMSTKLogManager
from ramstk.views.gtk3 import Gdk, GdkPixbuf, GObject, Gtk, _
from ramstk.views.gtk3.assistants import (
    CreateProject,
    ExportProject,
    ImportProject,
    OpenProject,
)
from ramstk.views.gtk3.books import RAMSTKModuleBook, RAMSTKWorkBook
from ramstk.views.gtk3.options import OptionsDialog
from ramstk.views.gtk3.preferences import PreferencesDialog
from ramstk.views.gtk3.widgets import RAMSTKMessageDialog

Tconfiguration = TypeVar(
    "Tconfiguration", RAMSTKUserConfiguration, RAMSTKSiteConfiguration
)


def destroy(__widget: Gtk.Widget, __event: Gdk.Event = None) -> None:
    """Quit the RAMSTK application.

    This function quits the RAMSTK application when the X in the upper right
    corner is pressed or if this function is called as a callback.

    :param __widget: the Gtk.Widget() that called this method.
    :param __event: the Gdk.Event() that called this method.
    :return: None
    :rtype: None
    """
    Gtk.main_quit()


class RAMSTKDesktop(Gtk.Window):
    """The base view for the RAMSTK Books.

    This is the container class for the List Book, Module Book, and Work Book.
    Attributes of the RAMSTKDesktop are:

    :cvar RAMSTK_USER_CONFIGURATION: the instance of the RAMSTK
        user configuration class.
    :type RAMSTK_USER_CONFIGURATION:
        :class:`ramstk.configuration.RAMSTKUserConfiguration`
    :cvar dict dic_books: dictionary holding a reference to each RAMSTK book.
    :cvar dict dic_tab_pos: dictionary holding the Gtk.PositionType()s for each
        of left, right, top, and bottom.
    :ivar int _n_screens: the number of monitors attached to the machine
        running RAMSTK.
    :ivar float _height: the height of the monitors attached to the machine
        running RAMSTK.
    :ivar float _width: the average width of each monitor attached to the
        machine running RAMSTK.
    :ivar menubar: the Gtk.MenuBar() for the RAMSTK ModuleBook menu.
    :type menubar: :class:`Gtk.MenuBar`
    :ivar progressbar: the Gtk.Progressbar() for displaying progress counters.
    :type progressbar: :class:`Gtk.Progressbar`
    :ivar statusbar: the Gtk.Statusbar() for displaying messages.
    :type statusbar: :class:`Gtk.Statusbar`
    :ivar toolbar: the Gtk.Toolbar() for the RAMSTK ModuleBook tools.
    :type toolbar: :class:`Gtk.Toolbar`
    """

    RAMSTK_SITE_CONFIGURATION: RAMSTKSiteConfiguration = RAMSTKSiteConfiguration()
    RAMSTK_USER_CONFIGURATION: RAMSTKUserConfiguration = RAMSTKUserConfiguration()

    def __init__(self, configuration: Tconfiguration, logger: RAMSTKLogManager) -> None:
        """Initialize an instance of the RAMSTK Book.

        :param configuration: a list containing the RAMSTK user
            configuration and RAMSTK site configuration class instances.
        :param logger: the RAMSTKLogManager class instance.
        """
        # pylint: disable=non-parent-init-called
        GObject.GObject.__init__(self)
        self.RAMSTK_USER_CONFIGURATION = configuration[0]  # type: ignore
        self.RAMSTK_SITE_CONFIGURATION = configuration[1]  # type: ignore

        # Initialize private dictionary attributes.

        # Initialize private list attributes.

        # Initialize private scalar attributes.
        self._logger = logger
        try:
            _screen: Gdk.Screen = Gdk.Screen()
            _display = _screen.get_default().get_display()
            _monitor = _display.get_monitor(0)
            self._height: float = _monitor.get_geometry().height
            self._n_screens: int = _display.get_n_monitors()
            self._width: float = _monitor.get_geometry().width
        except AttributeError:
            # When running on CI servers, there will be no monitor.  We also
            # don't need one.
            self._height = -1
            self._n_screens = 0
            self._width = -1

        # Initialize public dictionary attributes.

        # Initialize public list attributes.

        # Initialize public scalar attributes.
        self.menubar: Gtk.MenuBar = Gtk.MenuBar()
        self.progressbar: Gtk.ProgressBar = Gtk.ProgressBar()
        self.statusbar: Gtk.Statusbar = Gtk.Statusbar()
        self.toolbar: Gtk.Toolbar = Gtk.Toolbar()

        self.icoStatus: Gtk.StatusIcon = Gtk.StatusIcon()

        self.nbkModuleBook: RAMSTKModuleBook = RAMSTKModuleBook(
            self.RAMSTK_USER_CONFIGURATION, logger
        )
        self.nbkWorkBook: RAMSTKWorkBook = RAMSTKWorkBook(
            self.RAMSTK_USER_CONFIGURATION, logger
        )
        self.nbkWorkBook.RAMSTK_SITE_CONFIGURATION = self.RAMSTK_SITE_CONFIGURATION

        try:
            locale.setlocale(
                locale.LC_ALL, self.RAMSTK_USER_CONFIGURATION.RAMSTK_LOCALE
            )
        except locale.Error:
            locale.setlocale(locale.LC_ALL, "")

        self.__set_properties()
        self.__make_ui()
        self.__set_callbacks()

        # Subscribe to PyPubSub messages.
        pub.subscribe(
            self._on_request_open,
            "request_open_program",
        )
        pub.subscribe(
            self._on_select,
            "request_set_title",
        )
        pub.subscribe(
            self._do_set_status,
            "request_set_status",
        )
        pub.subscribe(
            self._do_raise_message_dialog,
            "do_log_critical_msg",
        )
        pub.subscribe(
            self._do_raise_message_dialog,
            "do_log_debug_msg",
        )
        pub.subscribe(
            self._do_raise_message_dialog,
            "do_log_error_msg",
        )
        pub.subscribe(
            self._do_raise_message_dialog,
            "do_log_warning_msg",
        )

    def _do_raise_message_dialog(self, logger_name: str, message: str) -> None:
        """Raise a message dialog to alert the user to an issue.

        :param logger_name: the name of the logger the message is indtended for.
            The message will only be presented to the user if it's the same or higher
            than the user selected log level.
        :param message: the info, debug, error, etc. message.
        :return: None
        :rtype: None
        """
        self.statusbar.push(2, message)
        if logger_name == self.RAMSTK_USER_CONFIGURATION.RAMSTK_LOGLEVEL:
            _dialog = RAMSTKMessageDialog(parent=self)
            _dialog.do_set_message(message)
            _dialog.do_set_message_type(logger_name.lower())
            if _dialog.do_run() == Gtk.ResponseType.OK:
                _dialog.do_destroy()

    @staticmethod
    def _do_request_close_project(__widget: Gtk.Widget) -> None:
        """Request to close the open RAMSTK Program.

        :param Gtk.Widget __widget: the Gtk.Widget() that called this method.
        :return: None
        :rtype: None
        """
        pub.sendMessage("request_close_project")

    def _do_request_options_assistant(self, __widget: Gtk.ImageMenuItem) -> None:
        """Request the EditOptions assistant be launched.

        :param __widget: the Gtk.ImageMenuItem() that called this class.
        :return: None
        :rtype: None
        """
        _dialog = OptionsDialog(parent=self)

        # See ISSUE #433
        if _dialog.do_run() == Gtk.ResponseType.OK:
            print("Need site admin or higher privileges.")

        _dialog.do_destroy()

    def _do_request_preferences_assistant(self, __widget: Gtk.ImageMenuItem) -> None:
        """Request the PreferencesDialog assistant be launched.

        :param __widget: the Gtk.ImageMenuItem() that called this class.
        :return: None
        :rtype: None
        """
        _assistant = Gtk.Window()
        _preferences = PreferencesDialog(self.RAMSTK_USER_CONFIGURATION, self._logger)

        _screen: Gdk.Screen = Gdk.Screen()
        _n_screens = _screen.get_default().get_n_monitors()
        _width = _screen.width() / _n_screens
        _height = _screen.height()

        _assistant.set_border_width(5)
        _assistant.set_default_size(_width - 450, (4 * _height / 7))
        _assistant.set_modal(True)
        _assistant.set_position(Gtk.WindowPosition.CENTER)
        _assistant.set_resizable(True)
        _assistant.set_transient_for(self)

        _assistant.add(_preferences)

        _assistant.show_all()

    # noinspection PyDeepBugsSwappedArgs
    def _do_request_save_project(self, widget: Gtk.Widget, end: bool = False) -> None:
        """Request to save the open RAMSTK Program.

        :param Gtk.Widget widget: the Gtk.Widget() that called this method.
        :keyword bool end: indicates whether or not to quit RAMSTK after saving
            the project.
        :return: None
        :rtype: None
        """
        _message = _("Saving Program Database {0:s}").format(
            self.RAMSTK_USER_CONFIGURATION.RAMSTK_PROG_INFO["database"]
        )
        self.statusbar.push(2, _message)

        pub.sendMessage("request_save_project")

        if end:
            destroy(widget)

    def _do_set_status(self, status: str) -> None:
        """Set the status message.

        :param status: the status message to display.
        :return: None
        :rtype: None
        """
        self.statusbar.push(1, status)

    def _do_set_status_icon(self, connected: bool = False) -> None:
        """Set the status icon in the system tay to indicate connection status.

        :param bool connected: whether or not RAMSTK is connected to a program
            database.
        :return: None
        :rtype: None
        """
        _pixbuf: GdkPixbuf.Pixbuf = GdkPixbuf.Pixbuf()

        if connected:
            _icon: str = (
                self.RAMSTK_USER_CONFIGURATION.RAMSTK_ICON_DIR
                + "/32x32/db-connected.png"
            )
            _tooltip: str = _(
                f"RAMSTK is connected to program database "
                f"{self.RAMSTK_USER_CONFIGURATION.RAMSTK_PROG_INFO['database']}."
            )
        else:
            _icon = (
                self.RAMSTK_USER_CONFIGURATION.RAMSTK_ICON_DIR
                + "/32x32/db-disconnected.png"
            )
            _tooltip = _("RAMSTK is not currently connected to a project database.")

        self.icoStatus.set_from_pixbuf(_pixbuf.new_from_file_at_size(_icon, 22, 22))
        self.icoStatus.set_tooltip_markup(_tooltip)

    def _on_button_press(self, __book: object, event: Gdk.EventButton) -> None:
        """Handle mouse clicks on the RAMSTKBook.

        :param __book: the RAMSTKBook that was 'clicked'.
        :param event: the Gdk.Event() that called this method (the
                      important attribute is which mouse button was clicked).

                      * 1 = left
                      * 2 = scrollwheel
                      * 3 = right
                      * 4 = forward
                      * 5 = backward
                      * 8 =
                      * 9 =

        :type event: :class:`Gdk.Event`
        :return: None
        :rtype: None
        """
        if event.button == 1:
            self.present()

    # pylint: disable=unused-argument
    # noinspection PyUnusedLocal
    def _on_request_open(self, program_db: object, database: Dict[str, str]) -> None:
        """Set the status bar and update the progress bar.

        :param program_db: the BaseDatabase used for the ProgramDB.  Unused in this
            method, but required as it is part of the message payload.
        :param database: the dict containing the database connection information.
        :return: None
        :rtype: None
        """
        _message = _(
            f"Opening Program Database {database['database']} on server "
            f"{database['host']}"
        )
        # noinspection PyDeepBugsSwappedArgs
        self.statusbar.push(1, _message)

        _message = _(f"Analyzing {database['database']} on server {database['host']}")
        # noinspection PyDeepBugsSwappedArgs
        self.statusbar.push(1, _message)
        self.set_title(_message)

    def _on_select(self, title: str) -> None:
        """Respond to load the Work View Gtk.Notebook() widgets.

        This method handles the results of the an individual module's
        _on_select() method.  It sets the title of the RAMSTK Work Book and
        raises an error dialog if needed.

        :return: None
        :rtype: None
        """
        try:
            self.set_title(title)
        except AttributeError as _error:
            self.RAMSTK_LOGGER.do_log_exception(__name__, _error)
            self.do_raise_dialog(severity="warning", user_msg=_error)

    def __make_menu(self) -> None:
        """Make the menu for the Module Book.

        :return: None
        :rtype: None
        """
        self.menubar.append(self.__make_menu_file())
        self.menubar.append(self.__make_menu_edit())
        self.menubar.append(self.__make_menu_tools())

        self.menubar.show_all()

    def __make_menu_edit(self) -> Gtk.MenuItem:
        """Make the Edit menu.

        :return: the Edit menu.
        :rtype: :class:`Gtk.MenuItem`
        """
        _menu = Gtk.Menu()

        _menu_items = self.__make_menu_items(["preferences"], ["_Preferences"])

        for _menu_item in _menu_items:
            _menu.append(_menu_item)

        _menu_items[0].connect("activate", self._do_request_preferences_assistant)

        _menu_item = Gtk.MenuItem(label=_("_Edit"), use_underline=True)
        _menu_item.set_submenu(_menu)

        return _menu_item

    def __make_menu_file(self) -> Gtk.MenuItem:
        """Make the File menu.

        :return: the File menu.
        :rtype: :class:`Gtk.MenuItem`
        """
        _menu = Gtk.Menu()

        _menu_items = self.__make_menu_items(
            ["new", "open", "import", "export", "save", "", "exit"],
            [
                "New _Program",
                "_Open",
                "_Import Project",
                "_Export Project",
                "_Save",
                "_Close",
                "E_xit",
            ],
        )

        for _menu_item in _menu_items:
            _menu.append(_menu_item)

        _menu_items[0].connect(
            "activate", CreateProject, self.RAMSTK_USER_CONFIGURATION, self
        )
        _menu_items[1].connect(
            "activate", OpenProject, self.RAMSTK_USER_CONFIGURATION, self
        )
        _menu_items[2].connect(
            "activate", ImportProject, self.RAMSTK_USER_CONFIGURATION, self
        )
        _menu_items[3].connect(
            "activate", ExportProject, self.RAMSTK_USER_CONFIGURATION, self
        )
        _menu_items[4].connect("activate", self._do_request_save_project)
        _menu_items[5].connect("activate", self._do_request_close_project)
        _menu_items[6].connect("activate", destroy)

        _menu_item = Gtk.MenuItem()
        _menu_item.set_label(_("_File"))
        _menu_item.set_property("use_underline", True)
        _menu_item.set_submenu(_menu)

        return _menu_item

    def __make_menu_items(
        self, icons: List[str], labels: List[str]
    ) -> List[Gtk.ImageMenuItem]:
        """Make a list of the menu items for a menu.

        :param icons: the list of icons to display on each menu item.
        :param labels: the list of label text for each menu item.
        :return: _menu_items; the list of menu items to include in the menu.
        """
        _menu_items: List[Gtk.ImageMenuItem] = []

        for _idx, _ico in enumerate(icons):
            if _ico != "":
                _icon = (
                    self.RAMSTK_USER_CONFIGURATION.RAMSTK_ICON_DIR
                    + "/16x16/"
                    + _ico
                    + ".png"
                )
            else:
                _icon = (
                    self.RAMSTK_USER_CONFIGURATION.RAMSTK_ICON_DIR
                    + "/16x16/default.png"
                )

            _menu_item = Gtk.ImageMenuItem()
            _image = Gtk.Image()
            _image.set_from_file(_icon)
            _menu_item.set_label(_(labels[_idx]))
            _menu_item.set_image(_image)
            _menu_item.set_property("use_underline", True)

            _menu_items.append(_menu_item)

        return _menu_items

    def __make_menu_tools(self) -> Gtk.MenuItem:
        """Make the Tools menu.

        :return: the Tools menu.
        :rtype: :class:`Gtk.MenuItem`
        """
        _menu = Gtk.Menu()

        _menu_items = self.__make_menu_items(["options"], ["_Options"])

        for _menu_item in _menu_items:
            _menu.append(_menu_item)

        _menu_items[0].connect("activate", self._do_request_options_assistant)

        _menu_item = Gtk.MenuItem(label=_("_Tools"), use_underline=True)
        _menu_item.set_submenu(_menu)

        return _menu_item

    def __make_toolbuttons(
        self, icons: List[str], tooltips: List[str]
    ) -> List[Gtk.ToolButton]:
        """Make the toolbuttons for a toolbar.

        :param icons: the list of icons to display on the toolbutton.
        :param tooltips: the list of tooltips to associate with each
            toolbutton.
        :return: _buttons; the list of toolbuttons to display in the toolbar.
        """
        _buttons: List[Gtk.ToolButton] = []

        for _idx, _ico in enumerate(icons):
            if _ico != "":
                _icon = (
                    self.RAMSTK_USER_CONFIGURATION.RAMSTK_ICON_DIR
                    + "/32x32/"
                    + _ico
                    + ".png"
                )
            else:
                _icon = (
                    self.RAMSTK_USER_CONFIGURATION.RAMSTK_ICON_DIR
                    + "/32x32/default.png"
                )

            _button = Gtk.ToolButton()
            _button.set_tooltip_text(tooltips[_idx])
            _image = Gtk.Image()
            _image.set_from_file(_icon)
            _button.set_icon_widget(_image)

            _buttons.append(_button)

        return _buttons

    def __make_toolbar(self) -> None:
        """Make the toolbar for the Module Book.

        :return: None
        :rtype: None
        """
        _toolbuttons = self.__make_toolbuttons(
            ["new", "open", "close", "save", "save-exit", "exit"],
            [
                _("Create a new RAMSTK Program Database."),
                _("Connect to an existing RAMSTK Program Database."),
                _("Closes the open RAMSTK Program Database."),
                _("Save the currently open RAMSTK Project Database."),
                _("Save the currently open RAMSTK Program Database then quits."),
                _("Quits without saving the currently open RAMSTK Program Database."),
            ],
        )

        for _position, _toolbutton in enumerate(_toolbuttons):
            self.toolbar.insert(_toolbutton, _position)

        self.toolbar.insert(Gtk.SeparatorToolItem(), 3)
        self.toolbar.insert(Gtk.SeparatorToolItem(), 5)

        _toolbuttons[0].connect(
            "clicked", CreateProject, self.RAMSTK_USER_CONFIGURATION, self
        )
        _toolbuttons[1].connect(
            "clicked", OpenProject, self.RAMSTK_USER_CONFIGURATION, self
        )
        _toolbuttons[2].connect("clicked", self._do_request_close_project)
        _toolbuttons[3].connect("clicked", self._do_request_save_project)
        _toolbuttons[4].connect("clicked", self._do_request_save_project, True)
        _toolbuttons[5].connect("clicked", destroy)

        self.toolbar.show_all()

    def __make_ui(self) -> None:
        """Build the user interface.

        :return: None
        :rtype: None
        """
        self.__make_menu()
        self.__make_toolbar()

        _vbox = Gtk.VBox()
        _vbox.pack_start(self.menubar, False, False, 0)
        _vbox.pack_start(self.toolbar, False, False, 0)

        _vpaned = Gtk.VPaned()
        _vpaned.pack1(self.nbkModuleBook, True, False)
        _vpaned.pack2(self.nbkWorkBook, True, False)

        _vbox.pack_start(_vpaned, True, True, 0)
        _vbox.pack_start(self.statusbar, False, False, 0)

        self.add(_vbox)
        self.show_all()

        self.nbkModuleBook.set_current_page(0)

        self.statusbar.push(1, _("Ready"))
        self._do_set_status_icon()

    def __set_callbacks(self) -> None:
        """Set the callbacks for the RAMSTKBook() and widgets.

        :return: None
        :rtype: None
        """
        self.connect("delete_event", destroy)
        self.connect("button_press_event", self._on_button_press)

    def __set_properties(self) -> None:
        """Set properties of the RAMSTKBook and widgets.

        :return: None
        :rtype: None
        """
        self.set_border_width(5)
        self.set_position(Gtk.WindowPosition.NONE)
        self.set_resizable(True)

        _width = 15 * self._width / 16
        _height = 15 * self._height / 16
        self.resize(_width, _height)
        self.move(50, 5)



# pylint: disable=duplicate-code
# -*- coding: utf-8 -*-
#
#       ramstk.models.dbrecords.programdb_mechanism_record.py is part of The RAMSTK
#       Project
#
# All rights reserved.
# Copyright since 2007 Doyle "weibullguy" Rowland doyle.rowland <AT> reliaqual <DOT> com
"""RAMSTKMechanism Record Model."""

# Third Party Imports
from sqlalchemy import Column, ForeignKeyConstraint, Integer, String
from sqlalchemy.orm import relationship

# RAMSTK Local Imports
from .. import RAMSTK_BASE
from .baserecord import RAMSTKBaseRecord


class RAMSTKMechanismRecord(RAMSTK_BASE, RAMSTKBaseRecord):  # type: ignore
    """Class to represent table ramstk_mechanism in RAMSTK Program database.

    This table shares a Many-to-One relationship with ramstk_mode. This table
    shares a One-to-Many relationship with ramstk_cause. This table shares a One-
    to-Many relationship with ramstk_op_load.
    """

    __defaults__ = {
        "description": "",
        "pof_include": 1,
        "rpn": 0,
        "rpn_detection": 10,
        "rpn_detection_new": 10,
        "rpn_new": 0,
        "rpn_occurrence": 10,
        "rpn_occurrence_new": 10,
    }
    __tablename__ = "ramstk_mechanism"
    __table_args__ = (
        ForeignKeyConstraint(
            ["fld_revision_id", "fld_hardware_id", "fld_mode_id"],
            [
                "ramstk_mode.fld_revision_id",
                "ramstk_mode.fld_hardware_id",
                "ramstk_mode.fld_mode_id",
            ],
        ),
        {"extend_existing": True},
    )

    revision_id = Column("fld_revision_id", Integer, primary_key=True, nullable=False)
    hardware_id = Column(
        "fld_hardware_id",
        Integer,
        primary_key=True,
        default=-1,
        nullable=False,
    )
    mode_id = Column("fld_mode_id", Integer, primary_key=True, nullable=False)
    mechanism_id = Column(
        "fld_mechanism_id",
        Integer,
        primary_key=True,
        autoincrement=True,
        nullable=False,
        unique=True,
    )

    description = Column(
        "fld_description", String(512), default=__defaults__["description"]
    )
    pof_include = Column(
        "fld_pof_include", Integer, default=__defaults__["pof_include"]
    )
    rpn = Column("fld_rpn", Integer, default=__defaults__["rpn"])
    rpn_detection = Column(
        "fld_rpn_detection", Integer, default=__defaults__["rpn_detection"]
    )
    rpn_detection_new = Column(
        "fld_rpn_detection_new",
        Integer,
        default=__defaults__["rpn_detection_new"],
    )
    rpn_new = Column("fld_rpn_new", Integer, default=__defaults__["rpn_new"])
    rpn_occurrence = Column(
        "fld_rpn_occurrence", Integer, default=__defaults__["rpn_occurrence"]
    )
    rpn_occurrence_new = Column(
        "fld_rpn_occurrence_new",
        Integer,
        default=__defaults__["rpn_occurrence_new"],
    )

    # Define the relationships to other tables in the RAMSTK Program database.
    mode: relationship = relationship(
        "RAMSTKModeRecord",
        back_populates="mechanism",
    )
    cause: relationship = relationship(
        "RAMSTKCauseRecord",
        back_populates="mechanism",
        cascade="all,delete",
    )
    op_load: relationship = relationship(
        "RAMSTKOpLoadRecord",
        back_populates="mechanism",
        cascade="all,delete",
    )

    is_mode = False
    is_mechanism = True
    is_cause = False
    is_control = False
    is_action = False
    is_opload = False
    is_opstress = False
    is_testmethod = False

    def get_attributes(self):
        """Retrieve the current values of the Mechanism data model attributes.

        :return: {mode_id, mechanism_id, description, pof_include, rpn,
                  rpn_detection, rpn_detection_new, rpn_new, rpn_occurrence,
                  rpn_occurrence_new} pairs
        :rtype: dict
        """
        return {
            "mode_id": self.mode_id,
            "mechanism_id": self.mechanism_id,
            "description": self.description,
            "pof_include": self.pof_include,
            "rpn": self.rpn,
            "rpn_detection": self.rpn_detection,
            "rpn_detection_new": self.rpn_detection_new,
            "rpn_new": self.rpn_new,
            "rpn_occurrence": self.rpn_occurrence,
            "rpn_occurrence_new": self.rpn_occurrence_new,
        }



# -*- coding: utf-8 -*-
#
#       ramstk.models.dbtables.programdb_nswc_table.py is part of The RAMSTK Project
#
# All rights reserved.
# Copyright since 2007 Doyle "weibullguy" Rowland doyle.rowland <AT> reliaqual <DOT> com
"""RAMSTKNSWC Table Model."""

# Standard Library Imports
from datetime import date
from typing import Dict, Type, Union

# Third Party Imports
import treelib
from pubsub import pub

# RAMSTK Local Imports
from ..dbrecords import RAMSTKNSWCRecord
from .basetable import RAMSTKBaseTable


class RAMSTKNSWCTable(RAMSTKBaseTable):
    """Contain attributes and methods of the NSWC table model."""

    # Define private dictionary class attributes.

    # Define private list class attributes.

    # Define private scalar class attributes.
    _db_id_colname = "fld_hardware_id"
    _db_tablename = "ramstk_nswc"
    _select_msg = "selected_revision"
    _tag = "nswc"

    # Define public dictionary class attributes.

    # Define public list class attributes.

    # Define public scalar class attributes.

    def __init__(self, **kwargs: Dict[str, Union[float, int, str]]) -> None:
        """Initialize a RAMSTKNSWC table model instance."""
        super().__init__(**kwargs)

        # Initialize private dictionary attributes.

        # Initialize private list attributes.
        self._lst_id_columns = [
            "revision_id",
            "hardware_id",
            "parent_id",
            "record_id",
        ]

        # Initialize private scalar attributes.
        # This is the record class associated with the table being modelled.
        self._record: Type[RAMSTKNSWCRecord] = RAMSTKNSWCRecord

        # Initialize public dictionary attributes.

        # Initialize public list attributes.

        # Initialize public scalar attributes.
        self.pkey = "hardware_id"

        # Subscribe to PyPubSub messages.
        pub.subscribe(
            self._on_insert_hardware,
            "succeed_insert_hardware",
        )

    def do_get_new_record(  # pylint: disable=method-hidden
        self, attributes: Dict[str, Union[date, float, int, str]]
    ) -> RAMSTKNSWCRecord:
        """Gets a new record instance with attributes set.

        :param attributes: the dict of attribute values to assign to the new record.
        :return: None
        :rtype: None
        """
        _new_record = self._record()
        _new_record.revision_id = attributes["revision_id"]
        _new_record.hardware_id = attributes["hardware_id"]

        return _new_record

    def _on_insert_hardware(self, tree: treelib.Tree) -> None:
        """Add new node to the NSWC tree for the newly added Hardware.

        NSWC records are added by triggers in the database when a new Hardware item is
        added.  This method simply adds a new node to the NSWC tree with a blank record.

        :param tree: the Hardware tree with the new node.
        :return: None
        :rtype: None
        """
        for _node in tree.all_nodes()[1:]:
            if not self.tree.contains(_node.identifier) and _node.data["hardware"].part:
                _attributes = {
                    "revision_id": _node.data["hardware"].revision_id,
                    "hardware_id": _node.data["hardware"].hardware_id,
                }
                _record = self.do_get_new_record(_attributes)
                self.tree.create_node(
                    tag=self._tag,
                    identifier=_node.data["hardware"].hardware_id,
                    parent=0,
                    data={self._tag: _record},
                )

                pub.sendMessage(
                    f"succeed_insert_{self._tag}",
                    tree=self.tree,
                )



# -*- coding: utf-8 -*-
#
#       ramstk.models.db.common_database.py is part of the RAMSTK Project
#
# All rights reserved.
# Copyright since 2007 Doyle "weibullguy" Rowland doyle.rowland <AT> reliaqual <DOT> com
"""The RAMSTK common database model."""

# Standard Library Imports
import gettext
from datetime import date, datetime, timedelta
from typing import Dict, Tuple, Union

# Third Party Imports
from pubsub import pub

# RAMSTK Package Imports
from ramstk.configuration import RAMSTKSiteConfiguration, RAMSTKUserConfiguration

# RAMSTK Local Imports
from ..dbrecords import (
    RAMSTKCategoryRecord,
    RAMSTKFailureModeRecord,
    RAMSTKGroupRecord,
    RAMSTKHazardsRecord,
    RAMSTKLoadHistoryRecord,
    RAMSTKManufacturerRecord,
    RAMSTKMeasurementRecord,
    RAMSTKMethodRecord,
    RAMSTKModelRecord,
    RAMSTKRPNRecord,
    RAMSTKSiteInfoRecord,
    RAMSTKStakeholdersRecord,
    RAMSTKStatusRecord,
    RAMSTKSubCategoryRecord,
    RAMSTKTypeRecord,
    RAMSTKUserRecord,
)
from .basedatabase import BaseDatabase, do_create_program_db

_ = gettext.gettext


class RAMSTKCommonDB:
    """The RAMSTK common database manager model.

    The attributes of a RAMSTK common database manager are:

    :ivar dic_tables: a dict containing the instances of all the database table
        models this program database model is managing.
    :ivar site_configuration: the RAMSTKSiteConfiguration instance associated with
        this common database model.
    :ivar common_dao: the BaseDatabase() object that will connect to the
        RAMSTK common database.
    """

    def __init__(self) -> None:
        """Initialize an instance of the RAMSTK common database model."""

        # Initialize private dictionary attributes.

        # Initialize private list attributes.

        # Initialize private scalar attributes.

        # Initialize public dictionary attributes.
        self.dic_tables: Dict[str, object] = {
            "options": object,
        }

        # Initialize public list attributes.

        # Initialize public scalar attributes.
        self.site_configuration: RAMSTKSiteConfiguration = RAMSTKSiteConfiguration()
        self.common_dao: BaseDatabase = BaseDatabase()

        # Subscribe to PyPubSub messages.
        pub.subscribe(self.do_create_common, "request_create_common")

    def do_create_common(
        self,
        common_db: BaseDatabase,
        database: Dict[str, str],
        license_file: str,
    ) -> None:
        """Create a new RAMSTK Common database.

        :param common_db: the BaseDatabase() that is to be used to create and connect to
            the new RAMSTK common database.
        :param database: a dict containing the database connection arguments.
        :param license_file: the absolute path to the license file.
        :return: None
        :rtype: None
        """
        with open(
            self.site_configuration.RAMSTK_SITE_DIR
            + f'/{database["dialect"]}_common_db.sql',
            "r",
            encoding="utf-8",
        ) as _sql_file:
            self.common_dao = common_db
            do_create_program_db(database, _sql_file)
            self.common_dao.do_connect(database)
            self.do_load_site_info(license_file)

            _yn = (
                input(_("Would you like to add a RAMSTK Administrator? ([y]/n): "))
                or "y"
            )
            if _yn.lower() == "y":
                self.do_add_administrator()

            self.common_dao.do_disconnect()

            pub.sendMessage(
                "succeed_create_common_database",
                common_db=self.common_dao,
                database=database,
            )

    def do_add_administrator(self) -> None:
        """Add a new administrator to the RAMSTK pool."""
        _user = RAMSTKUserRecord()
        _user.user_id = 0
        _user.user_group_id = 1

        _user.user_lname = input(  # nosec
            _("Enter the RAMSTK Administrator's last name (surname): ")
        )
        _user.user_fname = input(  # nosec
            _("Enter the RAMSTK Administrator's first name (given name): ")
        )
        _user.user_email = input(  # nosec
            _("Enter the RAMSTK Administrator's e-mail address: ")
        )
        _user.user_phone = input(  # nosec
            _("Enter the RAMSTK Administrator's phone number: ")
        )

        self.common_dao.session.add(_user)
        self.common_dao.session.commit()

    def do_load_site_info(self, license_file: str) -> None:
        """Load the Site Information table.

        :param license_file: the absolute path to the license file.
        :return: None
        :rtype: None
        """
        _dic_site_info: Dict[str, Union[date, int, str]] = {
            "site_name": "DEMO",
            "product_key": "DEMO",
            "expire_on": date.today() + timedelta(days=30),
            "function_enabled": 1,
            "requirement_enabled": 1,
            "hardware_enabled": 1,
            "software_enabled": 0,
            "rcm_enabled": 0,
            "testing_enabled": 0,
            "incident_enabled": 0,
            "survival_enabled": 0,
            "vandv_enabled": 1,
            "hazard_enabled": 1,
            "stakeholder_enabled": 1,
            "allocation_enabled": 1,
            "similar_item_enabled": 1,
            "fmea_enabled": 1,
            "pof_enabled": 1,
            "rbd_enabled": 0,
            "fta_enabled": 0,
        }
        _site_id: int = -1

        try:
            with open(
                license_file,
                "r",
                encoding="UTF-8",
            ) as _license_file:
                _contents = _license_file.readlines()
                _site_id = int(_contents[0].strip("\n"))
                _dic_site_info["product_key"] = _contents[1].strip("\n")
                _dic_site_info["expire_on"] = datetime.strptime(
                    _contents[2].strip("\n"), "%Y-%m-%d"
                )
                _dic_site_info["function_enabled"] = int(_contents[3].strip("\n"))
                _dic_site_info["requirement_enabled"] = int(_contents[4].strip("\n"))
                _dic_site_info["hardware_enabled"] = int(_contents[5].strip("\n"))
                _dic_site_info["software_enabled"] = int(_contents[6].strip("\n"))
                _dic_site_info["rcm_enabled"] = int(_contents[7].strip("\n"))
                _dic_site_info["testing_enabled"] = int(_contents[8].strip("\n"))
                _dic_site_info["incident_enabled"] = int(_contents[9].strip("\n"))
                _dic_site_info["survival_enabled"] = int(_contents[10].strip("\n"))
                _dic_site_info["vandv_enabled"] = int(_contents[11].strip("\n"))
                _dic_site_info["hazard_enabled"] = int(_contents[12].strip("\n"))
                _dic_site_info["stakeholder_enabled"] = int(_contents[13].strip("\n"))
                _dic_site_info["allocation_enabled"] = int(_contents[14].strip("\n"))
                _dic_site_info["similar_item_enabled"] = int(_contents[15].strip("\n"))
                _dic_site_info["fmea_enabled"] = int(_contents[16].strip("\n"))
                _dic_site_info["pof_enabled"] = int(_contents[17].strip("\n"))
                _dic_site_info["rbd_enabled"] = int(_contents[18].strip("\n"))
                _dic_site_info["fta_enabled"] = int(_contents[19].strip("\n"))
                _dic_site_info["site_name"] = _contents[20].strip("\n")
        except IOError:
            _error_msg = (
                "Unable to read license key file.  Defaulting to a 30-day demo license."
            )
            pub.sendMessage("fail_read_license", error_message=_error_msg)

        _site_info = RAMSTKSiteInfoRecord()
        _site_info.site_id = _site_id
        _site_info.set_attributes(_dic_site_info)

        self.common_dao.session.add(_site_info)
        self.common_dao.session.commit()

    def do_load_site_variables(
        self, user_configuration: RAMSTKUserConfiguration
    ) -> RAMSTKUserConfiguration:
        """Load the RAMSTKUserConfiguration global variables from the site db.

        :param user_configuration: the RAMSTKUserConfiguration instance whose
            variable is to be loaded.
        :return: user_configuration
        :rtype: RAMSTKUserConfiguration
        """
        pub.sendMessage(
            "do_log_info_msg",
            logger_name="INFO",
            message="Loading global RAMSTK configuration variables.",
        )

        user_configuration = self._do_load_action_categories(user_configuration)
        user_configuration = self._do_load_action_status(user_configuration)
        user_configuration = self._do_load_affinity_groups(user_configuration)
        user_configuration = self._do_load_damage_models(user_configuration)
        user_configuration = self._do_load_detection_methods(user_configuration)
        user_configuration = self._do_load_hardware_variables(user_configuration)
        user_configuration = self._do_load_hazards(user_configuration)
        user_configuration = self._do_load_incident_categories(user_configuration)
        user_configuration = self._do_load_incident_status(user_configuration)
        user_configuration = self._do_load_incident_types(user_configuration)
        user_configuration = self._do_load_load_history(user_configuration)
        user_configuration = self._do_load_manufacturers(user_configuration)
        user_configuration = self._do_load_measureable_parameters(user_configuration)
        user_configuration = self._do_load_measurement_units(user_configuration)
        user_configuration = self._do_load_requirement_types(user_configuration)
        user_configuration = self._do_load_rpn_detection(user_configuration)
        user_configuration = self._do_load_rpn_occurrence(user_configuration)
        user_configuration = self._do_load_rpn_severity(user_configuration)
        user_configuration = self._do_load_severity(user_configuration)
        user_configuration = self._do_load_stakeholders(user_configuration)
        user_configuration = self._do_load_users(user_configuration)
        user_configuration = self._do_load_validation_types(user_configuration)
        user_configuration = self._do_load_workgroups(user_configuration)

        pub.sendMessage(
            "do_log_info_msg",
            logger_name="INFO",
            message="Loaded global RAMSTK configuration variables.",
        )

        return user_configuration

    def _do_load_action_categories(
        self, user_configuration: RAMSTKUserConfiguration
    ) -> RAMSTKUserConfiguration:
        """Load the action category dict.

        :param user_configuration: the RAMSTKUserConfiguration instance whose
            variable is to be loaded.
        :return: user_configuration
        :rtype: RAMSTKUserConfiguration
        """
        for _record in (
            self.common_dao.session.query(RAMSTKCategoryRecord)
            .filter(RAMSTKCategoryRecord.category_type == "action")
            .all()
        ):
            _attributes = _record.get_attributes()
            user_configuration.RAMSTK_ACTION_CATEGORY[_record.category_id] = (
                _attributes["name"],
                _attributes["description"],
                _attributes["category_type"],
                _attributes["value"],
            )

        return user_configuration

    def _do_load_action_status(
        self,
        user_configuration: RAMSTKUserConfiguration,
    ) -> RAMSTKUserConfiguration:
        """Load the action status dict.

        :param user_configuration: the RAMSTKUserConfiguration instance whose
            variable is to be loaded.
        :return: user_configuration
        :rtype: RAMSTKUserConfiguration
        """
        for _record in (
            self.common_dao.session.query(RAMSTKStatusRecord)
            .filter(RAMSTKStatusRecord.status_type == "action")
            .all()
        ):
            _attributes = _record.get_attributes()
            user_configuration.RAMSTK_ACTION_STATUS[_record.status_id] = (
                _attributes["name"],
                _attributes["description"],
                _attributes["status_type"],
            )

        return user_configuration

    def _do_load_affinity_groups(
        self, user_configuration: RAMSTKUserConfiguration
    ) -> RAMSTKUserConfiguration:
        """Load the affinity group dict.

        :param user_configuration: the RAMSTKUserConfiguration instance whose
            variable is to be loaded.
        :return: user_configuration
        :rtype: RAMSTKUserConfiguration
        """
        for _record in (
            self.common_dao.session.query(RAMSTKGroupRecord)
            .filter(RAMSTKGroupRecord.group_type == "affinity")
            .all()
        ):
            _attributes = _record.get_attributes()
            user_configuration.RAMSTK_AFFINITY_GROUPS[_record.group_id] = (
                _attributes["description"],
                _attributes["group_type"],
            )

        return user_configuration

    def _do_load_damage_models(
        self, user_configuration: RAMSTKUserConfiguration
    ) -> RAMSTKUserConfiguration:
        """Load the damage model dict.

        :param user_configuration: the RAMSTKUserConfiguration instance whose
            variable is to be loaded.
        :return: user_configuration
        :rtype: RAMSTKUserConfiguration
        """
        for _record in (
            self.common_dao.session.query(RAMSTKModelRecord)
            .filter(RAMSTKModelRecord.model_type == "damage")
            .all()
        ):
            _attributes = _record.get_attributes()
            user_configuration.RAMSTK_DAMAGE_MODELS[_record.model_id] = _attributes[
                "description"
            ]

        return user_configuration

    def _do_load_detection_methods(
        self, user_configuration: RAMSTKUserConfiguration
    ) -> RAMSTKUserConfiguration:
        """Load the RPN detection methods dict.

        :param user_configuration: the RAMSTKUserConfiguration instance whose
            variable is to be loaded.
        :return: user_configuration
        :rtype: RAMSTKUserConfiguration
        """
        for _record in (
            self.common_dao.session.query(RAMSTKMethodRecord)
            .filter(RAMSTKMethodRecord.method_type == "detection")
            .all()
        ):
            _attributes = _record.get_attributes()
            user_configuration.RAMSTK_DETECTION_METHODS[_record.method_id] = (
                _attributes["name"],
                _attributes["description"],
                _attributes["method_type"],
            )

        return user_configuration

    def _do_load_failure_modes(
        self, category_id: int, subcategory_id: int
    ) -> Dict[int, Union[float, str]]:
        """Load dict of failure modes.

        :param category_id: the category ID for the failure modes to load.
        :param subcategory_id: the subcategory ID for the failure modes to load.
        :return: the dict of failures modes for the passed category ID, subcategory
            ID pair.
        :rtype: dict
        """
        return {
            _mode.mode_id: [  # type: ignore
                _mode.description,
                _mode.mode_ratio,
                _mode.source,
            ]
            for _mode in (
                self.common_dao.session.query(RAMSTKFailureModeRecord)
                .filter(RAMSTKFailureModeRecord.category_id == category_id)
                .filter(RAMSTKFailureModeRecord.subcategory_id == subcategory_id)
                .all()
            )
        }

    def _do_load_hardware_variables(
        self, user_configuration: RAMSTKUserConfiguration
    ) -> RAMSTKUserConfiguration:
        """Load variables associated with hardware categories and failure modes.

        :param user_configuration: the RAMSTKUserConfiguration instance whose
            variable is to be loaded.
        :return: user_configuration
        :rtype: RAMSTKUserConfiguration
        """
        for _record in (
            self.common_dao.session.query(RAMSTKCategoryRecord)
            .filter(RAMSTKCategoryRecord.category_type == "hardware")
            .all()
        ):

            _subcats = {}
            user_configuration.RAMSTK_FAILURE_MODES[_record.category_id] = {}
            user_configuration.RAMSTK_STRESS_LIMITS[
                _record.category_id
            ] = self._do_load_stress_limits(_record)

            for _subcat in (
                self.common_dao.session.query(RAMSTKSubCategoryRecord)
                .filter(RAMSTKSubCategoryRecord.category_id == _record.category_id)
                .all()
            ):
                _subcats[_subcat.subcategory_id] = _subcat.description

                user_configuration.RAMSTK_FAILURE_MODES[_record.category_id][
                    _subcat.subcategory_id
                ] = {}
                user_configuration.RAMSTK_FAILURE_MODES[_record.category_id][
                    _subcat.subcategory_id
                ] = self._do_load_failure_modes(
                    _record.category_id, _subcat.subcategory_id
                )

            user_configuration.RAMSTK_CATEGORIES[
                _record.category_id
            ] = _record.description
            user_configuration.RAMSTK_SUBCATEGORIES[_record.category_id] = _subcats

        return user_configuration

    def _do_load_hazards(
        self, user_configuration: RAMSTKUserConfiguration
    ) -> RAMSTKUserConfiguration:
        """Load the hazards dict.

        :param user_configuration: the RAMSTKUserConfiguration instance whose
            variable is to be loaded.
        :return: user_configuration
        :rtype: RAMSTKUserConfiguration
        """
        for _record in self.common_dao.session.query(RAMSTKHazardsRecord).all():
            _attributes = _record.get_attributes()
            user_configuration.RAMSTK_HAZARDS[_record.hazard_id] = (
                _attributes["hazard_category"],
                _attributes["hazard_subcategory"],
            )

        return user_configuration

    def _do_load_incident_categories(
        self,
        user_configuration: RAMSTKUserConfiguration,
    ) -> RAMSTKUserConfiguration:
        """Load the incident category dict.

        :param user_configuration: the RAMSTKUserConfiguration instance whose
            variable is to be loaded.
        :return: user_configuration
        :rtype: RAMSTKUserConfiguration
        """
        for _record in (
            self.common_dao.session.query(RAMSTKCategoryRecord)
            .filter(RAMSTKCategoryRecord.category_type == "incident")
            .all()
        ):
            _attributes = _record.get_attributes()
            user_configuration.RAMSTK_INCIDENT_CATEGORY[_record.category_id] = (
                _attributes["name"],
                _attributes["description"],
                _attributes["category_type"],
                _attributes["value"],
            )

        return user_configuration

    def _do_load_incident_status(
        self,
        user_configuration: RAMSTKUserConfiguration,
    ) -> RAMSTKUserConfiguration:
        """Load the incident status dict.

        :param user_configuration: the RAMSTKUserConfiguration instance whose
            variable is to be loaded.
        :return: user_configuration
        :rtype: RAMSTKUserConfiguration
        """
        for _record in (
            self.common_dao.session.query(RAMSTKStatusRecord)
            .filter(RAMSTKStatusRecord.status_type == "incident")
            .all()
        ):
            _attributes = _record.get_attributes()
            user_configuration.RAMSTK_INCIDENT_STATUS[_record.status_id] = (
                _attributes["name"],
                _attributes["description"],
                _attributes["status_type"],
            )

        return user_configuration

    def _do_load_incident_types(
        self, user_configuration: RAMSTKUserConfiguration
    ) -> RAMSTKUserConfiguration:
        """Load incident type dict.

        :param user_configuration: the RAMSTKUserConfiguration instance whose
            variable is to be loaded.
        :return: user_configuration
        :rtype: RAMSTKUserConfiguration
        """
        for _record in (
            self.common_dao.session.query(RAMSTKTypeRecord)
            .filter(RAMSTKTypeRecord.type_type == "incident")
            .all()
        ):
            _attributes = _record.get_attributes()
            user_configuration.RAMSTK_INCIDENT_TYPE[_record.type_id] = (
                _attributes["code"],
                _attributes["description"],
                _attributes["type_type"],
            )

        return user_configuration

    def _do_load_load_history(
        self, user_configuration: RAMSTKUserConfiguration
    ) -> RAMSTKUserConfiguration:
        """Load the load history dict.

        :param user_configuration: the RAMSTKUserConfiguration instance whose
            variable is to be loaded.
        :return: user_configuration
        :rtype: RAMSTKUserConfiguration
        """
        for _record in self.common_dao.session.query(RAMSTKLoadHistoryRecord).all():
            _attributes = _record.get_attributes()
            user_configuration.RAMSTK_LOAD_HISTORY[_record.history_id] = _attributes[
                "description"
            ]

        return user_configuration

    def _do_load_manufacturers(
        self, user_configuration: RAMSTKUserConfiguration
    ) -> RAMSTKUserConfiguration:
        """Load the manufacturer dict.

        :param user_configuration: the RAMSTKUserConfiguration instance whose
            variable is to be loaded.
        :return: user_configuration
        :rtype: RAMSTKUserConfiguration
        """
        for _record in self.common_dao.session.query(RAMSTKManufacturerRecord).all():
            _attributes = _record.get_attributes()
            user_configuration.RAMSTK_MANUFACTURERS[_record.manufacturer_id] = (
                _attributes["description"],
                _attributes["location"],
                _attributes["cage_code"],
            )

        return user_configuration

    def _do_load_measureable_parameters(
        self, user_configuration: RAMSTKUserConfiguration
    ) -> RAMSTKUserConfiguration:
        """Load the measureable parameters dict.

        :param user_configuration: the RAMSTKUserConfiguration instance whose
            variable is to be loaded.
        :return: user_configuration
        :rtype: RAMSTKUserConfiguration
        """
        for _record in (
            self.common_dao.session.query(RAMSTKMeasurementRecord)
            .filter(RAMSTKMeasurementRecord.measurement_type == "damage")
            .all()
        ):
            _attributes = _record.get_attributes()
            user_configuration.RAMSTK_MEASURABLE_PARAMETERS[_record.measurement_id] = (
                _attributes["code"],
                _attributes["description"],
                _attributes["measurement_type"],
            )

        return user_configuration

    def _do_load_measurement_units(
        self, user_configuration: RAMSTKUserConfiguration
    ) -> RAMSTKUserConfiguration:
        """Load the measurement units dict.

        :param user_configuration: the RAMSTKUserConfiguration instance whose
            variable is to be loaded.
        :return: user_configuration
        :rtype: RAMSTKUserConfiguration
        """
        for _record in (
            self.common_dao.session.query(RAMSTKMeasurementRecord)
            .filter(RAMSTKMeasurementRecord.measurement_type == "unit")
            .all()
        ):
            _attributes = _record.get_attributes()
            user_configuration.RAMSTK_MEASUREMENT_UNITS[_record.measurement_id] = (
                _attributes["code"],
                _attributes["description"],
                _attributes["measurement_type"],
            )

        return user_configuration

    def _do_load_requirement_types(
        self, user_configuration: RAMSTKUserConfiguration
    ) -> RAMSTKUserConfiguration:
        """Load the requirement type dict.

        :param user_configuration: the RAMSTKUserConfiguration instance whose
            variable is to be loaded.
        :return: user_configuration
        :rtype: RAMSTKUserConfiguration
        """
        for _record in (
            self.common_dao.session.query(RAMSTKTypeRecord)
            .filter(RAMSTKTypeRecord.type_type == "requirement")
            .all()
        ):
            _attributes = _record.get_attributes()
            user_configuration.RAMSTK_REQUIREMENT_TYPE[_record.type_id] = (
                _attributes["code"],
                _attributes["description"],
                _attributes["type_type"],
            )

        return user_configuration

    def _do_load_rpn_detection(
        self, user_configuration: RAMSTKUserConfiguration
    ) -> RAMSTKUserConfiguration:
        """Load the RPN detection dict.

        :param user_configuration: the RAMSTKUserConfiguration instance whose
            variable is to be loaded.
        :return: user_configuration
        :rtype: RAMSTKUserConfiguration
        """
        for _record in (
            self.common_dao.session.query(RAMSTKRPNRecord)
            .filter(RAMSTKRPNRecord.rpn_type == "detection")
            .all()
        ):
            user_configuration.RAMSTK_RPN_DETECTION[
                _record.value
            ] = _record.get_attributes()

        return user_configuration

    def _do_load_rpn_occurrence(
        self, user_configuration: RAMSTKUserConfiguration
    ) -> RAMSTKUserConfiguration:
        """Load the RPN occurremce dict.

        :param user_configuration: the RAMSTKUserConfiguration instance whose
            variable is to be loaded.
        :return: user_configuration
        :rtype: RAMSTKUserConfiguration
        """
        for _record in (
            self.common_dao.session.query(RAMSTKRPNRecord)
            .filter(RAMSTKRPNRecord.rpn_type == "occurrence")
            .all()
        ):
            user_configuration.RAMSTK_RPN_OCCURRENCE[
                _record.value
            ] = _record.get_attributes()

        return user_configuration

    def _do_load_rpn_severity(
        self, user_configuration: RAMSTKUserConfiguration
    ) -> RAMSTKUserConfiguration:
        """Load the RPN severity dict.

        :param user_configuration: the RAMSTKUserConfiguration instance whose
            variable is to be loaded.
        :return: user_configuration
        :rtype: RAMSTKUserConfiguration
        """
        for _record in (
            self.common_dao.session.query(RAMSTKRPNRecord)
            .filter(RAMSTKRPNRecord.rpn_type == "severity")
            .all()
        ):
            user_configuration.RAMSTK_RPN_SEVERITY[
                _record.value
            ] = _record.get_attributes()

        return user_configuration

    def _do_load_stakeholders(
        self, user_configuration: RAMSTKUserConfiguration
    ) -> RAMSTKUserConfiguration:
        """Load the stakeholders dict.

        :param user_configuration: the RAMSTKUserConfiguration instance whose
            variable is to be loaded.
        :return: user_configuration
        :rtype: RAMSTKUserConfiguration
        """
        for _record in self.common_dao.session.query(RAMSTKStakeholdersRecord).all():
            _attributes = _record.get_attributes()
            user_configuration.RAMSTK_STAKEHOLDERS[
                _record.stakeholders_id
            ] = _attributes["stakeholder"]

        return user_configuration

    def _do_load_severity(
        self, user_configuration: RAMSTKUserConfiguration
    ) -> RAMSTKUserConfiguration:
        """Load the RAMSTK_SEVERITY variable.

        :param user_configuration: the RAMSTKUserConfiguration instance whose
            variable is to be loaded.
        :return: user_configuration
        :rtype: RAMSTKUserConfiguration
        """
        for _record in (
            self.common_dao.session.query(RAMSTKCategoryRecord)
            .filter(RAMSTKCategoryRecord.category_type == "risk")
            .all()
        ):
            _attributes = _record.get_attributes()
            user_configuration.RAMSTK_SEVERITY[_record.category_id] = (
                _attributes["name"],
                _attributes["description"],
                _attributes["category_type"],
                _attributes["value"],
            )

        return user_configuration

    @staticmethod
    def _do_load_stress_limits(
        record: RAMSTKCategoryRecord,
    ) -> Tuple[float, float, float, float, float, float, float, float, float, float]:
        """Load the electrical stress limits for hardware categories.

        :param record: the RAMSTKCategoryRecord with the stress limit information.
        :return: tuple of stress limits.
        :rtype: tuple
        """
        return (
            record.harsh_ir_limit,
            record.mild_ir_limit,
            record.harsh_pr_limit,
            record.mild_pr_limit,
            record.harsh_vr_limit,
            record.mild_vr_limit,
            record.harsh_deltat_limit,
            record.mild_deltat_limit,
            record.harsh_maxt_limit,
            record.mild_maxt_limit,
        )

    def _do_load_users(
        self, user_configuration: RAMSTKUserConfiguration
    ) -> RAMSTKUserConfiguration:
        """Load the RAMSTK user dict.

        :param user_configuration: the RAMSTKUserConfiguration instance whose
            variable is to be loaded.
        :return: user_configuration
        :rtype: RAMSTKUserConfiguration
        """
        for _record in self.common_dao.session.query(RAMSTKUserRecord).all():
            _attributes = _record.get_attributes()
            user_configuration.RAMSTK_USERS[_record.user_id] = (
                _attributes["user_lname"],
                _attributes["user_fname"],
                _attributes["user_email"],
                _attributes["user_phone"],
                _attributes["user_group_id"],
            )

        return user_configuration

    def _do_load_validation_types(
        self, user_configuration: RAMSTKUserConfiguration
    ) -> RAMSTKUserConfiguration:
        """Load validation types dict.

        :param user_configuration: the RAMSTKUserConfiguration instance whose
            variable is to be loaded.
        :return: user_configuration
        :rtype: RAMSTKUserConfiguration
        """

        for _record in (
            self.common_dao.session.query(RAMSTKTypeRecord)
            .filter(RAMSTKTypeRecord.type_type == "validation")
            .all()
        ):
            _attributes = _record.get_attributes()
            user_configuration.RAMSTK_VALIDATION_TYPE[_record.type_id] = (
                _attributes["code"],
                _attributes["description"],
                _attributes["type_type"],
            )

        return user_configuration

    def _do_load_workgroups(
        self, user_configuration: RAMSTKUserConfiguration
    ) -> RAMSTKUserConfiguration:
        """Load the RAMSTK workgroup dict.

        :param user_configuration: the RAMSTKUserConfiguration instance whose
            variable is to be loaded.
        :return: user_configuration
        :rtype: RAMSTKUserConfiguration
        """
        for _record in (
            self.common_dao.session.query(RAMSTKGroupRecord)
            .filter(RAMSTKGroupRecord.group_type == "workgroup")
            .all()
        ):
            _attributes = _record.get_attributes()
            user_configuration.RAMSTK_WORKGROUPS[_record.group_id] = (
                _attributes["description"],
                _attributes["group_type"],
            )

        return user_configuration



# Standard Library Imports
from datetime import date
from typing import Any, Dict, List, Type, Union

# Third Party Imports
import pandas as pd

# RAMSTK Local Imports
from ..dbrecords import RAMSTKValidationRecord as RAMSTKValidationRecord
from .basetable import RAMSTKBaseTable as RAMSTKBaseTable

class RAMSTKValidationTable(RAMSTKBaseTable):
    _db_id_colname: str
    _db_tablename: str
    _select_msg: str
    _tag: str
    _dic_status: Dict[Any, float]
    _lst_id_columns: List[str]
    _record: Type[RAMSTKValidationRecord]
    pkey: str
    def __init__(self, **kwargs: Dict[Any, Union[float, int, str]]) -> None: ...
    def do_get_new_record(
        self, attributes: Dict[str, Union[date, float, int, str]]
    ) -> RAMSTKValidationRecord: ...
    def do_calculate_plan(self) -> None: ...
    def _do_calculate_all_tasks(self) -> None: ...
    def _do_calculate_task(self, node_id: int) -> None: ...
    def _do_select_assessment_targets(self) -> pd.DataFrame: ...



class RAMSTKError(Exception):
    def __init__(self, msg: str = ...) -> None: ...

class DataAccessError(RAMSTKError):
    msg: str
    def __init__(self, msg: str) -> None: ...

class OutOfRangeError(RAMSTKError):
    msg: str
    def __init__(self, msg: str) -> None: ...



# -*- coding: utf-8 -*-
#
#       ramstk.models.dbtables.basetable.py is part of The RAMSTK Project
#
# All rights reserved.
# Copyright since 2007 Doyle "weibullguy" Rowland doyle.rowland <AT> reliaqual <DOT> com
"""Metaclass for the database table models."""

# Standard Library Imports
from datetime import date
from typing import Any, Callable, Dict, List, Type, Union

# Third Party Imports
import treelib
from pubsub import pub
from treelib.exceptions import NodeIDAbsentError

# RAMSTK Package Imports
from ramstk.exceptions import DataAccessError
from ramstk.models.db import BaseDatabase
from ramstk.views.gtk3 import _


def do_clear_tree(tree: treelib.Tree) -> treelib.Tree:
    """Clear all nodes from a tree except the root node.

    :param tree: the treelib.Tree to clear.
    :return: the original tree with all child nodes removed.
    :rtype: treelib.Tree
    """
    for _node in tree.children(tree.root):
        tree.remove_node(_node.identifier)

    return tree


class RAMSTKBaseTable:
    """Metaclass for all RAMSTK Table Models.

    :cvar _db_id_colname: the name of the primary key column in the database.
    :cvar _db_tablename: the name of the database table.
    :cvar _root: the root node in the Treelib.tree; nominally 0.
    :cvar _select_msg: the message to listen for to call the do_select_all() method.
    :cvar _tag: the name of the RAMSTK work flow module.  This is the same for all
        classes associated with the work flow module.

    :ivar _lst_id_columns: the list of column names in the database table that
        contain an ID value.  These are generally primary and/or foreign key columns.
    :ivar _parent_id: the ID of the parent object in a hierarchical structure such as a
        hardware BoM.
    :ivar _record: the database record model object associated with the table model.
    :ivar _revision_id: the ID of the revision that is currently selected.
    :ivar dao: the instanace of the RAMSTK Program database model.
    :ivar last_id: the last ID number used in the database table.
    :ivar pkey: the name of the primary key column for the database table.
    :ivar tree: the treelib Tree()that will contain the structure of the RAMSTK
        module being modeled.
    """

    # Define private dictionary class attributes.

    # Define private list class attributes.

    # Define private scalar class attributes.
    _db_id_colname: str = ""
    _db_tablename: str = ""
    _root: int = 0
    _select_msg: str = "selected_revision"
    _tag: str = ""

    # Define public dictionary class attributes.

    # Define public list class attributes.

    # Define public scalar class attributes.

    # pylint: disable=unused-argument
    # noinspection PyUnusedLocal
    def __init__(self, **kwargs: Dict[str, Union[float, int, str]]) -> None:
        """Initialize an RAMSTK table model instance."""

        # Initialize private dictionary attributes.

        # Initialize private list attributes.
        self._lst_id_columns: List[str] = []

        # Initialize private scalar attributes.
        self._parent_id: int = 0
        self._record: Type[object]
        self._revision_id: int = 0

        # Initialize public dictionary attributes.

        # Initialize public list attributes.

        # Initialize public scalar attributes.
        self.dao: BaseDatabase = BaseDatabase()
        self.last_id: int = 0
        self.pkey: str = ""
        self.tree: treelib.Tree = treelib.Tree()
        self.do_get_new_record: Callable[
            [Dict[str, Union[date, float, int, str]]], object
        ]

        # Add the root to the Tree().  This is necessary to allow multiple
        # entries at the top level as there can only be one root in a treelib
        # Tree().  Manipulation and viewing of a RAMSTK module tree needs to
        # ignore the root of the tree.
        self.tree.create_node(tag=self._tag, identifier=self._root)

        # Subscribe to PyPubSub messages.
        pub.subscribe(self.do_connect, "succeed_connect_program_database")
        pub.subscribe(self.do_delete, f"request_delete_{self._tag}")
        pub.subscribe(self.do_get_attributes, f"request_get_{self._tag}_attributes")
        pub.subscribe(self.do_get_tree, f"request_get_{self._tag}_tree")
        pub.subscribe(self.do_insert, f"request_insert_{self._tag}")
        pub.subscribe(self.do_select_all, self._select_msg)
        pub.subscribe(self.do_set_attributes, f"request_set_{self._tag}_attributes")
        pub.subscribe(self.do_set_attributes, f"lvw_editing_{self._tag}")
        pub.subscribe(self.do_set_attributes, f"mvw_editing_{self._tag}")
        pub.subscribe(self.do_set_attributes, f"wvw_editing_{self._tag}")
        pub.subscribe(
            self.do_set_attributes_all,
            f"request_set_all_{self._tag}_attributes",
        )
        pub.subscribe(self.do_set_tree, f"succeed_calculate_{self._tag}")
        pub.subscribe(self.do_update, f"request_update_{self._tag}")
        pub.subscribe(self.do_update_all, f"request_update_all_{self._tag}")
        pub.subscribe(self.do_update_all, "request_save_project")

    def do_connect(self, dao: BaseDatabase) -> None:
        """Connect data manager to a database.

        :param dao: the BaseDatabase() instance (data access object)
            representing the connected RAMSTK Program database.
        :type dao: :class:`ramstk.db.base.BaseDatabase`
        """
        self.dao = dao

    def do_create_all_codes(self, prefix: str) -> None:
        """Create codes for all MODULE data table records.

        :param prefix: the string to use as a prefix for each code.
        :return: None
        :rtype: None
        """
        for _node in self.tree.all_nodes():
            # noinspection PyUnresolvedReferences
            self.do_create_code(_node.identifier, prefix)  # type: ignore

    def do_delete(self, node_id: int) -> None:  # sourcery skip: extract-method
        """Remove a record from the Program database and records tree.

        :param node_id: the ID of the record to delete.
        :return: None
        :rtype: None
        """
        try:
            for _node in self.tree.children(node_id):
                _record = self.do_select(_node.identifier)
                self.dao.do_delete(_record)

            _record = self.do_select(node_id)
            self.dao.do_delete(_record)

            self.tree.remove_node(node_id)
            self.last_id = self.dao.get_last_id(self._db_tablename, self._db_id_colname)

            pub.sendMessage(
                f"succeed_delete_{self._tag}",
                tree=self.tree,
            )
        except (AttributeError, DataAccessError, NodeIDAbsentError):
            pub.sendMessage(
                "do_log_debug_msg",
                logger_name="DEBUG",
                message=_(
                    f"Attempted to delete non-existent "
                    f"{self._tag.replace('_', ' ').title()} ID {node_id}."
                ),
            )

    def do_get_attributes(self, node_id: int) -> None:
        """Retrieve the RAMSTK data table attributes for node ID.

        :param node_id: the node ID in the treelib Tree to get the attributes for.
        :return: None
        :rtype: None
        """
        try:
            pub.sendMessage(
                f"succeed_get_{self._tag}_attributes",
                attributes=self.do_select(node_id).get_attributes(),
            )
        except AttributeError:
            pub.sendMessage(
                "do_log_debug_msg",
                logger_name="DEBUG",
                message=_(
                    f"No attributes found for record ID {node_id} in {self._tag} table."
                ),
            )

    def do_get_tree(self) -> None:
        """Retrieve the records tree.

        :return: None
        :rtype: None
        """
        pub.sendMessage(
            f"succeed_get_{self._tag}_tree",
            tree=self.tree,
        )

    def do_insert(
        self, attributes: Dict[str, Union[date, float, int, str]]
    ) -> None:  # sourcery skip: extract-method
        """Add a new record to the RAMSTK program database and records tree.

        :param attributes: the attribute values to assign to the new record.
        :return: None
        :rtype: None
        """
        try:
            _record = self.do_get_new_record(attributes)
            for _id in self._lst_id_columns:
                attributes.pop(_id)

            _record.set_attributes(attributes)  # type: ignore
            _identifier = self.last_id + 1

            self.dao.do_insert(_record)
            self.last_id = self.dao.get_last_id(self._db_tablename, self._db_id_colname)

            self.tree.create_node(
                tag=self._tag,
                identifier=_identifier,
                parent=self._parent_id,
                data={self._tag: _record},
            )

            pub.sendMessage(
                f"succeed_insert_{self._tag}",
                tree=self.tree,
            )
        except DataAccessError as _error:
            pub.sendMessage(
                "do_log_debug_msg",
                logger_name="DEBUG",
                message=str(_error.msg),
            )
        except NodeIDAbsentError as _error:
            pub.sendMessage(
                "do_log_debug_msg",
                logger_name="DEBUG",
                message=str(_error),
            )

    def do_select(self, node_id: Any) -> Any:
        """Retrieve the RAMSTK data table record for the Node ID passed.

        :param node_id: the Node ID of the data package to retrieve.
        :return: the instance of the RAMSTK<MODULE> data table that was
            requested or None if the requested Node ID does not exist.
        :raise: KeyError if passed the name of a table that isn't managed by
            this manager.
        """
        try:
            _entity = self.tree.get_node(node_id).data[self._tag]
        except (
            AttributeError,
            KeyError,
            treelib.tree.NodeIDAbsentError,
            TypeError,
        ):
            pub.sendMessage(
                "do_log_debug_msg",
                logger_name="DEBUG",
                message=_(
                    f"No data package for node ID {node_id} in module {self._tag}."
                ),
            )
            _entity = None

        return _entity

    def do_select_all(self, attributes: Dict[str, Union[float, int, str]]) -> None:
        """Retrieve all the records from the RAMSTK Program database.

        :param attributes: the attribute dict for the selected record.
        :return: None
        :rtype: None
        """
        self.tree = do_clear_tree(self.tree)

        _keys = [_key for _key in self._lst_id_columns if _key in attributes]
        _values = [
            attributes[_key]  # type: ignore
            for _key in self._lst_id_columns
            if _key in attributes
        ]

        for _record in self.dao.do_select_all(
            self._record,
            key=_keys,
            value=_values,
            order=self._db_id_colname,
        ):
            try:
                self._parent_id = _record.get_attributes()["parent_id"]
            except KeyError:
                self._parent_id = 0

            self.tree.create_node(
                tag=self._tag,
                identifier=_record.get_attributes()[self.pkey],
                parent=self._parent_id,
                data={self._tag: _record},
            )
        self.last_id = self.dao.get_last_id(self._db_tablename, self._db_id_colname)

        pub.sendMessage(
            f"succeed_retrieve_all_{self._tag}",
            tree=self.tree,
        )

    def do_set_attributes(
        self, node_id: int, package: Dict[str, Union[float, int, str]]
    ) -> None:
        """Set the attributes of the record associated with node ID.

        :param node_id: the ID of the record in the RAMSTK Program database
            table whose attributes are to be set.
        :param package: the key:value pair of the attribute to set.
        :return: None
        :rtype: None
        """
        [[_key, _value]] = package.items()

        try:
            _attributes = self.do_select(node_id).get_attributes()
        except (AttributeError, KeyError):
            pub.sendMessage(
                "do_log_debug_msg",
                logger_name="DEBUG",
                message=_(
                    f"No data package for node ID {node_id} in module {self._tag}."
                ),
            )
            _attributes = {}

        for _field in self._lst_id_columns:
            try:
                _attributes.pop(_field)
            except KeyError:
                pass

        if _key in _attributes:
            _attributes[_key] = _value

            self.do_select(node_id).set_attributes(_attributes)

        # noinspection PyUnresolvedReferences
        self.do_get_tree()  # type: ignore

    def do_set_attributes_all(
        self, attributes: Dict[str, Union[float, int, str]]
    ) -> None:
        """Set all the attributes of the record associated with the Module ID.

        :param attributes: the aggregate attribute dict for the allocation item.
        :return: None
        :rtype: None
        """
        for _key in attributes:
            self.do_set_attributes(
                node_id=attributes[self.pkey],  # type: ignore
                package={_key: attributes[_key]},
            )

    def do_set_tree(self, tree: treelib.Tree) -> None:
        """Set the MODULE treelib Tree().

        This method is generally used to respond to events such as successful
        calculations of the entire system.

        :param tree: the treelib Tree() to assign to the tree attribute.
        :return: None
        :rtype: None
        """
        self.tree = tree

    def do_update(self, node_id: int) -> None:
        """Update record associated with node ID in RAMSTK Program database.

        :param node_id: the node ID of the record to save.
        :return: None
        :rtype: None
        """
        try:
            self.dao.do_update(self.tree.get_node(node_id).data[self._tag])
            pub.sendMessage(
                f"succeed_update_{self._tag}",
                tree=self.tree,
            )
        except AttributeError:
            pub.sendMessage(
                "do_log_debug_msg",
                logger_name="DEBUG",
                message=_(
                    f"Attempted to save non-existent {self._tag.replace('_', ' ')} "
                    f"with {self._tag.replace('_', ' ')} ID {node_id}."
                ),
            )
        except KeyError:
            pub.sendMessage(
                "do_log_debug_msg",
                logger_name="DEBUG",
                message=_(
                    f"No data package found for {self._tag.replace('_', ' ')} ID "
                    f"{node_id}."
                ),
            )
        except (DataAccessError, TypeError):
            if node_id != 0:
                _error_msg = _(
                    f"The value for one or more attributes for "
                    f"{self._tag.replace('_', ' ')} ID {node_id} was the wrong type."
                )
            else:
                _error_msg = _(f"Attempting to update the root node {node_id}.")
            pub.sendMessage(
                "do_log_debug_msg",
                logger_name="DEBUG",
                message=_error_msg,
            )

        pub.sendMessage("request_set_cursor_active")

    # noinspection PyUnresolvedReferences
    # pylint: disable=no-value-for-parameter
    def do_update_all(self) -> None:
        """Update all MODULE data table records in the RAMSTK Program database.

        :return: None
        :rtype: None
        """
        for _node in self.tree.all_nodes():
            self.do_update(_node.identifier)  # type: ignore

        pub.sendMessage("request_set_cursor_active")
        pub.sendMessage(f"succeed_update_all_{self._tag}")



# pylint: disable=unused-import
# -*- coding: utf-8 -*-
#
#       ramstk.models.dbtables.__init__.py is part of The RAMSTK Project
#
# All rights reserved.
# Copyright since 2007 Doyle "weibullguy" Rowland doyle.rowland <AT> reliaqual <DOT> com
"""The RAMSTK database table models package."""

# RAMSTK Local Imports
from .commondb_category_table import RAMSTKCategoryTable
from .commondb_condition_table import RAMSTKConditionTable
from .commondb_failure_mode_table import RAMSTKFailureModeTable
from .commondb_group_table import RAMSTKGroupTable
from .commondb_hazards_table import RAMSTKHazardsTable
from .commondb_load_history_table import RAMSTKLoadHistoryTable
from .commondb_manufacturer_table import RAMSTKManufacturerTable
from .commondb_measurement_table import RAMSTKMeasurementTable
from .commondb_method_table import RAMSTKMethodTable
from .commondb_model_table import RAMSTKModelTable
from .commondb_rpn_table import RAMSTKRPNTable
from .commondb_site_info_table import RAMSTKSiteInfoTable
from .commondb_stakeholders_table import RAMSTKStakeholdersTable
from .commondb_status_table import RAMSTKStatusTable
from .commondb_subcategory_table import RAMSTKSubCategoryTable
from .commondb_type_table import RAMSTKTypeTable
from .commondb_user_table import RAMSTKUserTable
from .programdb_action_table import RAMSTKActionTable
from .programdb_allocation_table import RAMSTKAllocationTable
from .programdb_cause_table import RAMSTKCauseTable
from .programdb_control_table import RAMSTKControlTable
from .programdb_design_electric_table import RAMSTKDesignElectricTable
from .programdb_design_mechanic_table import RAMSTKDesignMechanicTable
from .programdb_environment_table import RAMSTKEnvironmentTable
from .programdb_failure_definition_table import RAMSTKFailureDefinitionTable
from .programdb_function_table import RAMSTKFunctionTable
from .programdb_hardware_table import RAMSTKHardwareTable
from .programdb_hazard_table import RAMSTKHazardTable
from .programdb_mechanism_table import RAMSTKMechanismTable
from .programdb_milhdbk217f_table import RAMSTKMILHDBK217FTable
from .programdb_mission_phase_table import RAMSTKMissionPhaseTable
from .programdb_mission_table import RAMSTKMissionTable
from .programdb_mode_table import RAMSTKModeTable
from .programdb_nswc_table import RAMSTKNSWCTable
from .programdb_opload_table import RAMSTKOpLoadTable
from .programdb_opstress_table import RAMSTKOpStressTable
from .programdb_program_info_table import RAMSTKProgramInfoTable
from .programdb_program_status_table import RAMSTKProgramStatusTable
from .programdb_reliability_table import RAMSTKReliabilityTable
from .programdb_requirement_table import RAMSTKRequirementTable
from .programdb_revision_table import RAMSTKRevisionTable
from .programdb_similar_item_table import RAMSTKSimilarItemTable
from .programdb_stakeholder_table import RAMSTKStakeholderTable
from .programdb_test_method_table import RAMSTKTestMethodTable
from .programdb_validation_table import RAMSTKValidationTable
